# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for generating maccor procedure files from
input parameters and procedure templates
"""

import os
import time
from copy import deepcopy

import numpy as np
import pandas as pd
import xmltodict
from beep.protocol import PROCEDURE_TEMPLATE_DIR, PROTOCOL_SCHEMA_DIR
from beep.conversion_schemas import MACCOR_WAVEFORM_CONFIG
from beep.utils import DashOrderedDict
from beep.utils.waveform import convert_velocity_to_power_waveform, RapidChargeWave


class Procedure(DashOrderedDict):
    """
    Procedure file object. Provides factory methods
    to read a Maccor-type procedure file and invoke
    from templates for specific experimental
    procedure parameters

    """

    @classmethod
    def from_file(cls, filename, encoding="UTF-8"):
        """
        Procedure file ingestion. Invokes Procedure object
        from standard Maccor xml file.

        Args:
            filename (str): xml procedure file.

        Returns:
            (Procedure): Ordered dictionary with keys corresponding to options or
                control variables. Section headers are nested dicts or lists
                within the dict.
        """
        with open(filename, "rb") as f:
            text = f.read().decode(encoding)
        data = xmltodict.parse(text, process_namespaces=False, strip_whitespace=True)
        return cls(data)

    # TODO: check on the necessity of this with MACCOR instrument
    def _format_maccor(self):
        """
        Dictionary reformatting of the entries in the procedure in
        order to match the maccor formats. Mainly re-adding whitespace
        to entries that were stripped on injestion.

        Returns:
            dict: Ordered dictionary with reformatted entries to match the
                formatting used in the maccor procedure files.
        """
        formatted = deepcopy(self)
        for step in formatted["MaccorTestProcedure"]["ProcSteps"]["TestStep"]:
            # print(json.dumps(step['StepType'], indent=2))
            while len(step["StepType"]) < 8:
                step["StepType"] = step["StepType"].center(8)
            if step["StepMode"] is None:
                step["StepMode"] = " "
            while len(step["StepMode"]) < 8:
                step["StepMode"] = step["StepMode"].center(8)
            if step["Ends"] is not None:
                # If the Ends Element is a list we need to
                # check each entry in the list
                if isinstance(step["Ends"]["EndEntry"], list):
                    # print(json.dumps(step['Ends'], indent=2))
                    for end_entry in step["Ends"]["EndEntry"]:
                        self.ends_whitespace(end_entry)
                if isinstance(step["Ends"]["EndEntry"], dict):
                    self.ends_whitespace(step["Ends"]["EndEntry"])
            if step["Reports"] is not None:
                if isinstance(step["Reports"]["ReportEntry"], list):
                    for rep_entry in step["Reports"]["ReportEntry"]:
                        self.reports_whitespace(rep_entry)
                if isinstance(step["Reports"]["ReportEntry"], dict):
                    self.reports_whitespace(step["Reports"]["ReportEntry"])

        return formatted

    @staticmethod
    def ends_whitespace(end_entry):
        if end_entry["SpecialType"] is None:
            end_entry["SpecialType"] = " "
        while len(end_entry["EndType"]) < 8:
            end_entry["EndType"] = end_entry["EndType"].center(8)
        if end_entry["Oper"] is not None:
            if len(end_entry["Oper"]) < 2:
                end_entry["Oper"] = end_entry["Oper"].center(3)
            else:
                end_entry["Oper"] = end_entry["Oper"].ljust(3)

    @staticmethod
    def reports_whitespace(rep_entry):
        while len(rep_entry["ReportType"]) < 8:
            rep_entry["ReportType"] = rep_entry["ReportType"].center(8)

    def to_file(self, filename, encoding="UTF-8"):
        """
        Writes object to maccor-formatted xml file using xmltodict
        unparse function.

        filename (str): full path and name to save the output
        encoding (str): text encoding of output file

        Args:
            filename (str):file name to save xml to.
        """
        formatted = self._format_maccor()
        contents = xmltodict.unparse(
            input_dict=formatted,
            output=None,
            encoding=encoding,
            short_empty_elements=False,
            pretty=True,
            newl="\n",
            indent="  ",
        )

        # Manually inject processing instructions on line 2
        line0, remainder = contents.split("\n", 1)
        line1 = '<?maccor-application progid="Maccor Procedure File"?>'
        contents = "\n".join([line0, line1, remainder])
        contents = self.fixup_empty_elements(contents)
        contents += "\n"
        with open(filename, "w") as f:
            f.write(contents)

    @staticmethod
    def fixup_empty_elements(text):
        """
        xml reformatting to match the empty elements that are used
        in the maccor procedure format. Writes directly back to the file
        and assumes that the empty elements to be replaced are all on a
        single line.

        Args:
            text (str): xml file raw text to be formatted

        """
        text = text.replace(r"<Limits></Limits>", "<Limits/>")
        text = text.replace(r"<Reports></Reports>", "<Reports/>")
        text = text.replace(r"<Ends></Ends>", "<Ends/>")
        return text

    def modify_step_value(self, step_num, step_type, step_value):
        """
        Modifies the procedure parameters to set a step value at at given step num and type.

        Args:
            step_num (int): step id at which to set value
            step_type (str): step type at which to set value
            step_value (str): value to set

        Returns:
            dict: modified proc_dict with set value
        """
        for step_idx, step in enumerate(
            self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        ):
            if step_idx == step_num and step["StepType"] == step_type:
                step["StepValue"] = step_value
        return self

    @classmethod
    def from_exp(cls, cutoff_voltage, charge_rate, discharge_rate, template=None):
        """
        Generates a procedure according to the EXP-style template.

        Args:
            cutoff_voltage (float): cutoff voltage for.
            charge_rate (float): charging C-rate in 1/h.
            discharge_rate (float): discharging C-rate in 1/h.
            template (str): template name, defaults to EXP in template dir

        Returns:
            (Procedure): dictionary of procedure parameters.

        """
        # Load EXP template
        template = template or os.path.join(PROCEDURE_TEMPLATE_DIR, "EXP.000")
        obj = cls.from_file(template)

        # Modify according to params
        loop_idx_start, loop_idx_end = None, None
        for step_idx, step in enumerate(
            obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        ):
            if step["StepType"] == "Do 1":
                loop_idx_start = step_idx
            if step["StepType"] == "Loop 1":
                loop_idx_end = step_idx

        if loop_idx_start is None or loop_idx_end is None:
            raise UnboundLocalError("Loop index is not set")

        for step_idx, step in enumerate(
            obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        ):
            if step["StepType"] == "Charge":
                if step["Limits"] is not None and "Voltage" in step["Limits"]:
                    step["Limits"]["Voltage"] = cutoff_voltage
                if (
                    step["StepMode"] == "Current"
                    and loop_idx_start < step_idx < loop_idx_end
                ):
                    step["StepValue"] = charge_rate
            if (
                step["StepType"] == "Dischrge"
                and step["StepMode"] == "Current"
                and loop_idx_start < step_idx < loop_idx_end
            ):
                step["StepValue"] = discharge_rate

        return obj

    # TODO: rename this diagnosticv2 and merge
    @classmethod
    def from_regcyclev2(cls, reg_param, template=None):
        """
        Generates a procedure according to the diagnosticV2 template.

        Args:
            reg_param (pandas.Dataframe): containing the following quantities
                charge_constant_current_1 (float): C
                charge_percent_limit_1 (float): % of nominal capacity
                charge_constant_current_2 (float): C
                charge_cutoff_voltage (float): V
                charge_constant_voltage_time (integer): mins
                charge_rest_time (integer): mins
                discharge_constant_current (float): C
                discharge_cutoff_voltage (float): V
                discharge_rest_time (integer): mins
                cell_temperature_nominal (float): ˚C
                capacity_nominal (float): Ah
                diagnostic_start_cycle (integer): cycles
                diagnostic_interval (integer): cycles

        Returns:
            (Procedure): dictionary of procedure parameters.
        """

        assert (
            reg_param["charge_cutoff_voltage"] > reg_param["discharge_cutoff_voltage"]
        )

        dc_idx = 1

        # Load template
        template = template or os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV2.000")
        obj = cls.from_file(template)
        obj.insert_resistance_regcyclev2(dc_idx, reg_param)

        # Start of initial set of regular cycles
        reg_charge_idx = 27 + 1
        obj.insert_charge_regcyclev2(reg_charge_idx, reg_param)
        reg_discharge_idx = 27 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Start of main loop of regular cycles
        reg_charge_idx = 59 + 1
        obj.insert_charge_regcyclev2(reg_charge_idx, reg_param)
        reg_discharge_idx = 59 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Storage cycle
        reg_storage_idx = 69
        obj.insert_storage_regcyclev2(reg_storage_idx, reg_param)

        return obj

    # TODO: These should probably all be private
    def insert_maccor_waveform(self, waveform_idx, waveform_filename):
        """
        Inserts a waveform into procedure dictionary at given id.

        Args:
            waveform_idx (int): Step in the procedure file to
                insert waveform at
            waveform_filename (str): Path to .MWF waveform file.
                Waveform needs to be pre-scaled for current/power
                capabilities of the cell and cycler

        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.StepType".format(waveform_idx),
            "FastWave",
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.StepMode".format(waveform_idx),
            "",
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Ends".format(waveform_idx), None
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Reports".format(waveform_idx),
            None,
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Range".format(waveform_idx), ""
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Option1".format(waveform_idx), ""
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Option2".format(waveform_idx), ""
        )
        self.set(
            "MaccorTestProcedure.ProcSteps.TestStep.{}.Option3".format(waveform_idx), ""
        )

        assert steps[waveform_idx]["StepType"] == "FastWave"
        assert waveform_filename.split(".")[-1].lower() == "mwf"
        local_name = waveform_filename.split(".")[0]
        _, local_name = os.path.split(local_name)
        assert len(local_name) < 25, str(len(local_name))

        steps[waveform_idx]["StepValue"] = local_name

        return self

    def insert_resistance_regcyclev2(self, resist_idx, reg_param):
        """
        Inserts resistance into procedure dictionary at given id.

        Args:
            resist_idx (int):
            reg_param (pandas.DataFrame):

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # Initial resistance check
        assert steps[resist_idx]["StepType"] == "Charge"
        assert steps[resist_idx]["StepMode"] == "Current"
        steps[resist_idx]["StepValue"] = float(
            round(1.0 * reg_param["capacity_nominal"], 3)
        )

        return self

    def insert_charge_regcyclev2(self, charge_idx, reg_param):
        """
        Inserts charge into procedure dictionary at given id.

        Args:
            charge_idx (int)
            reg_param (pandas.DataFrame):

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # Regular cycle constant current charge part 1
        step_idx = charge_idx
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(
                reg_param["charge_constant_current_1"] * reg_param["capacity_nominal"],
                3,
            )
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(
            round(
                3600
                * (reg_param["charge_percent_limit_1"] / 100)
                / reg_param["charge_constant_current_1"]
            )
        )
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Regular cycle constant current charge part 2
        step_idx = charge_idx + 1
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(
                reg_param["charge_constant_current_2"] * reg_param["capacity_nominal"],
                3,
            )
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = float(
            round(reg_param["charge_cutoff_voltage"], 3)
        )

        # Regular cycle constant voltage hold
        step_idx = charge_idx + 2
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Voltage"
        steps[step_idx]["StepValue"] = reg_param["charge_cutoff_voltage"]
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * reg_param["charge_constant_voltage_time"]))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Regular cycle rest at top of charge
        step_idx = charge_idx + 3
        assert steps[step_idx]["StepType"] == "Rest"
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * reg_param["charge_rest_time"]))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        return self

    def insert_discharge_regcyclev2(self, discharge_idx, reg_param):
        """
        Inserts discharge into procedure dictionary at given id.

        Args:
            discharge_idx (int):
            reg_param (pandas.DataFrame):

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # Regular cycle constant current discharge part 1
        step_idx = discharge_idx
        assert steps[step_idx]["StepType"] == "Dischrge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(
                reg_param["discharge_constant_current"] * reg_param["capacity_nominal"],
                3,
            )
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = float(
            round(reg_param["discharge_cutoff_voltage"], 3)
        )

        # Regular cycle rest after discharge
        step_idx = discharge_idx + 1
        assert steps[step_idx]["StepType"] == "Rest"
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * reg_param["discharge_rest_time"]))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Regular cycle number of times to repeat regular cycle for initial offset and main body
        step_idx = discharge_idx + 3
        assert steps[step_idx]["StepType"][0:4] == "Loop"
        if steps[step_idx]["StepType"] == "Loop 1":
            assert steps[step_idx]["Ends"]["EndEntry"]["EndType"] == "Loop Cnt"
            steps[step_idx]["Ends"]["EndEntry"]["Value"] = reg_param[
                "diagnostic_start_cycle"
            ]
        elif steps[step_idx]["StepType"] == "Loop 2":
            assert steps[step_idx]["Ends"]["EndEntry"]["EndType"] == "Loop Cnt"
            steps[step_idx]["Ends"]["EndEntry"]["Value"] = reg_param[
                "diagnostic_interval"
            ]

        return self

    def insert_storage_regcyclev2(self, storage_idx, reg_param):
        """
        Inserts storage into procedure dictionary at given id.

        Args:
            storage_idx (int):
            reg_param (pandas.DataFrame):

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # Storage condition
        step_idx = storage_idx
        assert steps[step_idx]["StepType"] == "Dischrge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(0.5 * reg_param["capacity_nominal"], 3)
        )
        steps[step_idx]["Limits"]["Voltage"] = float(
            round(reg_param["discharge_cutoff_voltage"], 3)
        )
        assert steps[step_idx]["Ends"]["EndEntry"]["EndType"] == "Current"
        steps[step_idx]["Ends"]["EndEntry"]["Value"] = float(
            round(0.05 * reg_param["capacity_nominal"], 3)
        )
        step_idx = storage_idx + 1
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(0.5 * reg_param["capacity_nominal"], 3)
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * 12))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        return self

    @classmethod
    def generate_procedure_regcyclev3(cls, protocol_index, reg_param, template=None):
        """
        Generates a procedure according to the diagnosticV3 template.

        Args:
            protocol_index (int): number of the protocol file being
                generated from this file.
            reg_param (pandas.DataFrame): containing the following quantities
                charge_constant_current_1 (float): C
                charge_percent_limit_1 (float): % of nominal capacity
                charge_constant_current_2 (float): C
                charge_cutoff_voltage (float): V
                charge_constant_voltage_time (integer): mins
                charge_rest_time (integer): mins
                discharge_constant_current (float): C
                discharge_cutoff_voltage (float): V
                discharge_rest_time (integer): mins
                cell_temperature_nominal (float): ˚C
                capacity_nominal (float): Ah
                diagnostic_start_cycle (integer): cycles
                diagnostic_interval (integer): cycles
            template (str): template for invocation, defaults to
                the diagnosticV3.000 template

        Returns:
            (Procedure): dictionary invoked using template/parameters
        """
        assert (
            reg_param["charge_cutoff_voltage"] > reg_param["discharge_cutoff_voltage"]
        )

        rest_idx = 0

        template = template or os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV3.000")
        obj = cls.from_file(template)
        obj.insert_initialrest_regcyclev3(rest_idx, protocol_index)

        dc_idx = 1
        obj.insert_resistance_regcyclev2(dc_idx, reg_param)

        # Start of initial set of regular cycles
        reg_charge_idx = 27 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        reg_discharge_idx = 27 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Start of main loop of regular cycles
        reg_charge_idx = 59 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        reg_discharge_idx = 59 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Storage cycle
        reg_storage_idx = 93
        obj.insert_storage_regcyclev3(reg_storage_idx, reg_param)

        # Check that the upper charge voltage is set the same for both charge current steps
        reg_charge_idx = 27 + 1
        assert (
            obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx]["Ends"][
                "EndEntry"
            ][1]["Value"]
            == obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                "Ends"
            ]["EndEntry"][0]["Value"]
        )

        reg_charge_idx = 59 + 1
        assert (
            obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx]["Ends"][
                "EndEntry"
            ][1]["Value"]
            == obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                "Ends"
            ]["EndEntry"][0]["Value"]
        )

        return obj

    @classmethod
    def generate_procedure_drivingv1(cls, protocol_index, reg_param, waveform_name, template=None):
        """
        Generates a procedure according to the diagnosticV3 template.

        Args:
            protocol_index (int): number of the protocol file being
                generated from this file.
            reg_param (pandas.DataFrame): containing the following quantities
                charge_constant_current_1 (float): C
                charge_percent_limit_1 (float): % of nominal capacity
                charge_constant_current_2 (float): C
                charge_cutoff_voltage (float): V
                charge_constant_voltage_time (integer): mins
                charge_rest_time (integer): mins
                discharge_profile (str): {'US06', 'LA4', '9Lap'}
                profile_charge_limit (float): upper limit voltage for the profile
                max_profile_power (float): maximum power setpoint during the profile
                n_repeats (int): number of repetitions for the profile
                discharge_cutoff_voltage (float): V
                power_scaling (float): Power relative to the other profiles
                discharge_rest_time (integer): mins
                cell_temperature_nominal (float): ˚C
                capacity_nominal (float): Ah
                diagnostic_start_cycle (integer): cycles
                diagnostic_interval (integer): cycles
            waveform_name (str): Name of the waveform file (not path) without extension
            template (str): template for invocation, defaults to
                the diagnosticV3.000 template

        Returns:
            (Procedure): dictionary invoked using template/parameters
        """
        assert (
                reg_param["charge_cutoff_voltage"] > reg_param["discharge_cutoff_voltage"]
        )
        assert (
                reg_param["charge_constant_current_1"]
                <= reg_param["charge_constant_current_2"]
        )

        rest_idx = 0

        template = template or os.path.join(PROCEDURE_TEMPLATE_DIR, "drivingV1.000")
        obj = cls.from_file(template)
        obj.insert_initialrest_regcyclev3(rest_idx, protocol_index)

        dc_idx = 1
        obj.insert_resistance_regcyclev2(dc_idx, reg_param)

        # Start of initial set of regular cycles
        reg_charge_idx = 27 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        reg_discharge_idx = 27 + 5
        obj.insert_maccor_waveform(reg_discharge_idx, waveform_name)

        # Start of main loop of regular cycles
        reg_charge_idx = 59 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        reg_discharge_idx = 59 + 5
        obj.insert_maccor_waveform(reg_discharge_idx, waveform_name)

        # Storage cycle
        reg_storage_idx = 93
        obj.insert_storage_regcyclev3(reg_storage_idx, reg_param)

        # Check that the upper charge voltage is set the same for both charge current steps
        reg_charge_idx = 27 + 1
        assert (
                obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx]["Ends"][
                    "EndEntry"
                ][1]["Value"]
                == obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                    "Ends"
                ]["EndEntry"][0]["Value"]
        )

        reg_charge_idx = 59 + 1
        assert (
                obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx]["Ends"][
                    "EndEntry"
                ][1]["Value"]
                == obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                    "Ends"
                ]["EndEntry"][0]["Value"]
        )

        return obj

    @classmethod
    def generate_procedure_chargingv1(cls, protocol_index, reg_param, waveform_name, template=None):
        """
        Generates a procedure according to the diagnosticV4 template.

        Args:
            protocol_index (int): number of the protocol file being
                generated from this file.
            reg_param (pandas.DataFrame): containing the following quantities
                charge_type_1 (str): {'smooth', 'step'} type of charging waveform
                charge_start_soc_1 (float): assumed starting soc for the charge
                charge_current_param_1 (float): c-rate for first charging window
                charge_current_param_2 (float): c-rate for second charging window
                charge_current_param_3 (float): c-rate for third charging window
                charge_current_param_4 (float): c-rate for fourth charging window
                charge_soc_param_1 (float): soc point for dividing first charging window from second window
                charge_soc_param_2 (float): soc point for dividing second charging window from third window
                charge_fast_soc_limit (float): % of nominal capacity to end fast charging
                charge_cutoff_voltage (float): upper voltage limit for the charge
                charge_constant_voltage_time (integer): mins
                charge_rest_time (integer): mins
                profile_charge_limit (float): upper limit voltage for the profile
                max_profile_power (float): maximum power setpoint during the profile
                n_repeats (int): number of repetitions for the profile
                discharge_cutoff_voltage (float): V
                power_scaling (float): Power relative to the other profiles
                discharge_rest_time (integer): mins
                cell_temperature_nominal (float): ˚C
                capacity_nominal (float): Ah
                diagnostic_start_cycle (integer): cycles
                diagnostic_interval (integer): cycles
            waveform_name (str): Name of the waveform file (not path) without extension
            template (str): template for invocation, defaults to
                the diagnosticV4.000 template

        Returns:
            (Procedure): dictionary invoked using template/parameters
        """
        assert (
                reg_param["charge_cutoff_voltage"] > reg_param["discharge_cutoff_voltage"]
        )

        rest_idx = 0

        template = template or os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV4.000")
        obj = cls.from_file(template)
        obj.insert_initialrest_regcyclev3(rest_idx, protocol_index)

        dc_idx = 1
        obj.insert_resistance_regcyclev2(dc_idx, reg_param)

        # Set variables necessary to use regcyclev3 function
        reg_param["charge_constant_current_1"] = 1
        reg_param["charge_constant_current_2"] = reg_param["charge_current_param_4"]
        reg_param["charge_percent_limit_1"] = reg_param["charge_fast_soc_limit"]

        # Start of initial set of regular cycles
        reg_charge_idx = 27 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        obj.insert_maccor_waveform(reg_charge_idx, waveform_name)
        reg_discharge_idx = 27 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Start of main loop of regular cycles
        reg_charge_idx = 59 + 1
        obj.insert_charge_regcyclev3(reg_charge_idx, reg_param)
        obj.insert_maccor_waveform(reg_charge_idx, waveform_name)
        reg_discharge_idx = 59 + 5
        obj.insert_discharge_regcyclev2(reg_discharge_idx, reg_param)

        # Storage cycle
        reg_storage_idx = 93
        obj.insert_storage_regcyclev3(reg_storage_idx, reg_param)

        # Check that the upper charge voltage is lower for the fast charge portion
        reg_charge_idx = 27 + 1
        df_mwf = pd.read_csv(
            waveform_name,
            sep="\t",
            header=None,
        )
        waveform_charge_limit = df_mwf.loc[:, 4].max()

        assert (
                waveform_charge_limit
                < obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                    "Ends"
                ]["EndEntry"][0]["Value"]
        )

        reg_charge_idx = 59 + 1
        assert (
                waveform_charge_limit
                < obj["MaccorTestProcedure"]["ProcSteps"]["TestStep"][reg_charge_idx + 1][
                    "Ends"
                ]["EndEntry"][0]["Value"]
        )

        return obj

    def insert_initialrest_regcyclev3(self, rest_idx, index):
        """
        Inserts initial rest into procedure dictionary at given id.

        Args:
            rest_idx (int):
            index (int):

        Returns:
            dict:

        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        # Initial rest
        offset_seconds = 720
        assert steps[rest_idx]["StepType"] == "Rest"
        assert steps[rest_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(3 * 3600 + offset_seconds * (index % 96)))
        steps[rest_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        return self

    def insert_charge_regcyclev3(self, charge_idx, reg_param):
        """
        Inserts charge into procedure dictionary at given id.

        Args:
            charge_idx (int):
            reg_param (pandas.DataFrame):

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # Regular cycle constant current charge part 1
        step_idx = charge_idx
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(
                reg_param["charge_constant_current_1"] * reg_param["capacity_nominal"],
                3,
            )
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(
            round(
                3600
                * (reg_param["charge_percent_limit_1"] / 100)
                / reg_param["charge_constant_current_1"]
            )
        )
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )
        assert steps[step_idx]["Ends"]["EndEntry"][1]["Value"] != 4.4
        steps[step_idx]["Ends"]["EndEntry"][1]["Value"] = float(
            round(reg_param["charge_cutoff_voltage"], 3)
        )

        # Regular cycle constant current charge part 2
        step_idx = charge_idx + 1
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(
                reg_param["charge_constant_current_2"] * reg_param["capacity_nominal"],
                3,
            )
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = float(
            round(reg_param["charge_cutoff_voltage"], 3)
        )

        # Regular cycle constant voltage hold
        step_idx = charge_idx + 2
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Voltage"
        steps[step_idx]["StepValue"] = reg_param["charge_cutoff_voltage"]
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * reg_param["charge_constant_voltage_time"]))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Regular cycle rest at top of charge
        step_idx = charge_idx + 3
        assert steps[step_idx]["StepType"] == "Rest"
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * reg_param["charge_rest_time"]))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        return self

    def insert_storage_regcyclev3(self, storage_idx, reg_param):
        """
        Inserts storage into procedure dictionary at given id.

        Args:
            self (dict):
            storage_idx (int):
            reg_param (pandas.DataFrame):

        Returns:
            dict:

        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        # Storage condition
        step_idx = storage_idx
        assert steps[step_idx]["StepType"] == "Dischrge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(0.5 * reg_param["capacity_nominal"], 3)
        )
        steps[step_idx]["Limits"]["Voltage"] = float(
            round(reg_param["discharge_cutoff_voltage"], 3)
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = float(
            round(0.05 * reg_param["capacity_nominal"], 3)
        )
        step_idx = storage_idx + 1
        assert steps[step_idx]["StepType"] == "Charge"
        assert steps[step_idx]["StepMode"] == "Current"
        steps[step_idx]["StepValue"] = float(
            round(0.5 * reg_param["capacity_nominal"], 3)
        )
        assert steps[step_idx]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * 12))
        steps[step_idx]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        return self

    # TODO: should this just be part of the general template?
    def add_procedure_diagcyclev2(self, nominal_capacity, diagnostic_params):
        """
        Modifies a procedure according to the diagnosticV2 template.

        Args:
            nominal_capacity (float): Standard capacity for this cell.
            diagnostic_params (pandas.Series): Series containing all of the
                diagnostic parameters.

        Returns:
            (dict) dictionary of procedure parameters.

        """
        start_reset_cycle_1 = 4
        self.insert_reset_cyclev2(
            start_reset_cycle_1, nominal_capacity, diagnostic_params
        )
        start_hppc_cycle_1 = 8
        self.insert_hppc_cyclev2(
            start_hppc_cycle_1, nominal_capacity, diagnostic_params
        )
        start_rpt_cycle_1 = 18
        self.insert_rpt_cyclev2(start_rpt_cycle_1, nominal_capacity, diagnostic_params)

        start_reset_cycle_2 = 37
        self.insert_reset_cyclev2(
            start_reset_cycle_2, nominal_capacity, diagnostic_params
        )
        start_hppc_cycle_2 = 40
        self.insert_hppc_cyclev2(
            start_hppc_cycle_2, nominal_capacity, diagnostic_params
        )
        start_rpt_cycle_2 = 50
        self.insert_rpt_cyclev2(start_rpt_cycle_2, nominal_capacity, diagnostic_params)

        return self

    # TODO: make private
    def generate_procedure_diagcyclev3(self, nominal_capacity, diagnostic_params):
        """
        Generates a diagnostic procedure according
        to the diagnosticV3 template.

        Args:
            nominal_capacity (float): Standard capacity for this cell.
            diagnostic_params (pandas.Series): Series containing all of the
                diagnostic parameters.

        Returns:
            dict: dictionary of procedure parameters.

        """
        start_reset_cycle_1 = 4
        self.insert_reset_cyclev2(
            start_reset_cycle_1, nominal_capacity, diagnostic_params
        )
        start_hppc_cycle_1 = 8
        self.insert_hppc_cyclev2(
            start_hppc_cycle_1, nominal_capacity, diagnostic_params
        )
        start_rpt_cycle_1 = 18
        self.insert_rpt_cyclev2(start_rpt_cycle_1, nominal_capacity, diagnostic_params)

        start_reset_cycle_2 = 37
        self.insert_reset_cyclev2(
            start_reset_cycle_2, nominal_capacity, diagnostic_params
        )
        start_hppc_cycle_2 = 40
        self.insert_hppc_cyclev2(
            start_hppc_cycle_2, nominal_capacity, diagnostic_params
        )
        start_rpt_cycle_2 = 50
        self.insert_rpt_cyclev2(start_rpt_cycle_2, nominal_capacity, diagnostic_params)

        start_reset_cycle_3 = 70
        self.insert_reset_cyclev2(
            start_reset_cycle_3, nominal_capacity, diagnostic_params
        )
        start_hppc_cycle_3 = 74
        self.insert_hppc_cyclev2(
            start_hppc_cycle_3, nominal_capacity, diagnostic_params
        )
        start_rpt_cycle_3 = 84
        self.insert_rpt_cyclev2(start_rpt_cycle_3, nominal_capacity, diagnostic_params)

        return self

    def insert_reset_cyclev2(self, start, nominal_capacity, diagnostic_params):
        """
        Helper function for parameterizing the reset cycle.

        Args:
            start:
            nominal_capacity:
            diagnostic_params:

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        # Charge step for reset cycle
        assert steps[start]["StepType"] == "Charge"
        assert steps[start]["StepMode"] == "Current"
        steps[start]["StepValue"] = float(
            round(nominal_capacity * diagnostic_params["reset_cycle_current"], 4)
        )
        steps[start]["Limits"]["Voltage"] = float(
            round(diagnostic_params["diagnostic_charge_cutoff_voltage"], 3)
        )
        assert steps[start]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["reset_cycle_cutoff_current"], 4)
        )

        # Discharge step for reset cycle
        assert steps[start + 1]["StepType"] == "Dischrge"
        assert steps[start + 1]["StepMode"] == "Current"
        steps[start + 1]["StepValue"] = float(
            round(nominal_capacity * diagnostic_params["reset_cycle_current"], 4)
        )
        assert steps[start + 1]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[start + 1]["Ends"]["EndEntry"][0]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        return self

    def insert_hppc_cyclev2(self, start, nominal_capacity, diagnostic_params):
        """
        Helper function for parameterizing the hybrid pulse power cycle

        Args:
            start (int): index of the step to start at
            nominal_capacity (float): nominal capacity of the cell for calculating c-rate
            diagnostic_params (dict): dictionary containing parameters for the diagnostic cycles

        Returns:
            dict:
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        # Initial charge step for hppc cycle
        assert steps[start]["StepType"] == "Charge"
        assert steps[start]["StepMode"] == "Current"
        steps[start]["StepValue"] = float(
            round(
                nominal_capacity * diagnostic_params["HPPC_baseline_constant_current"],
                3,
            )
        )
        steps[start]["Limits"]["Voltage"] = float(
            round(diagnostic_params["diagnostic_charge_cutoff_voltage"], 3)
        )
        assert steps[start]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["diagnostic_cutoff_current"], 3)
        )

        # Rest step for hppc cycle
        assert steps[start + 2]["StepType"] == "Rest"
        assert steps[start + 2]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(60 * diagnostic_params["HPPC_rest_time"]))
        steps[start + 2]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Discharge step 1 for hppc cycle
        assert steps[start + 3]["StepType"] == "Dischrge"
        assert steps[start + 3]["StepMode"] == "Current"
        steps[start + 3]["StepValue"] = float(
            round(nominal_capacity * diagnostic_params["HPPC_pulse_current_1"], 3)
        )
        assert steps[start + 3]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = diagnostic_params["HPPC_pulse_duration_1"]
        steps[start + 3]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )
        assert steps[start + 3]["Ends"]["EndEntry"][1]["EndType"] == "Voltage"
        steps[start + 3]["Ends"]["EndEntry"][1]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        # Pulse rest step for hppc cycle
        assert steps[start + 4]["StepType"] == "Rest"
        assert steps[start + 4]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(round(diagnostic_params["HPPC_pulse_rest_time"]))
        steps[start + 4]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )

        # Charge step 1 for hppc cycle
        assert steps[start + 5]["StepType"] == "Charge"
        assert steps[start + 5]["StepMode"] == "Current"
        steps[start + 5]["StepValue"] = float(
            round(nominal_capacity * diagnostic_params["HPPC_pulse_current_2"], 3)
        )
        assert steps[start + 5]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = diagnostic_params["HPPC_pulse_duration_2"]
        steps[start + 5]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )
        assert steps[start + 5]["Ends"]["EndEntry"][1]["EndType"] == "Voltage"
        steps[start + 5]["Ends"]["EndEntry"][1]["Value"] = float(
            round(diagnostic_params["HPPC_pulse_cutoff_voltage"], 3)
        )

        # Discharge step 2 for hppc cycle
        assert steps[start + 6]["StepType"] == "Dischrge"
        assert steps[start + 6]["StepMode"] == "Current"
        steps[start + 6]["StepValue"] = float(
            round(
                nominal_capacity * diagnostic_params["HPPC_baseline_constant_current"],
                3,
            )
        )
        assert steps[start + 6]["Ends"]["EndEntry"][0]["EndType"] == "StepTime"
        time_s = int(
            round(
                3600
                * (diagnostic_params["HPPC_interval"] / 100)
                / diagnostic_params["HPPC_baseline_constant_current"]
                - 1
            )
        )
        steps[start + 6]["Ends"]["EndEntry"][0]["Value"] = time.strftime(
            "%H:%M:%S", time.gmtime(time_s)
        )
        assert steps[start + 6]["Ends"]["EndEntry"][1]["EndType"] == "Voltage"
        steps[start + 6]["Ends"]["EndEntry"][1]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        # Final discharge step for hppc cycle
        assert steps[start + 8]["StepType"] == "Dischrge"
        assert steps[start + 8]["StepMode"] == "Voltage"
        steps[start + 8]["StepValue"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )
        steps[start + 8]["Limits"]["Current"] = float(
            round(
                nominal_capacity * diagnostic_params["HPPC_baseline_constant_current"],
                3,
            )
        )
        assert steps[start + 8]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start + 8]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["diagnostic_cutoff_current"], 3)
        )

        return self

    def insert_rpt_cyclev2(self, start, nominal_capacity, diagnostic_params):
        """
        Helper function for parameterizing the Rate Performance Test cycle

        Args:
            start:
            nominal_capacity:
            diagnostic_params:

        Returns:
            dict
        """
        steps = self["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        # First charge step for rpt cycle
        assert steps[start]["StepType"] == "Charge"
        assert steps[start]["StepMode"] == "Current"
        steps[start]["StepValue"] = float(
            round(
                nominal_capacity * diagnostic_params["RPT_charge_constant_current"], 3
            )
        )
        steps[start]["Limits"]["Voltage"] = float(
            round(diagnostic_params["diagnostic_charge_cutoff_voltage"], 3)
        )
        assert steps[start]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["diagnostic_cutoff_current"], 3)
        )

        # 0.2C discharge step for rpt cycle
        assert steps[start + 1]["StepType"] == "Dischrge"
        assert steps[start + 1]["StepMode"] == "Current"
        steps[start + 1]["StepValue"] = float(
            round(
                nominal_capacity
                * diagnostic_params["RPT_discharge_constant_current_1"],
                3,
            )
        )
        assert steps[start + 1]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[start + 1]["Ends"]["EndEntry"][0]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        # Second charge step for rpt cycle
        assert steps[start + 3]["StepType"] == "Charge"
        assert steps[start + 3]["StepMode"] == "Current"
        steps[start + 3]["StepValue"] = float(
            round(
                nominal_capacity * diagnostic_params["RPT_charge_constant_current"], 3
            )
        )
        steps[start + 3]["Limits"]["Voltage"] = float(
            round(diagnostic_params["diagnostic_charge_cutoff_voltage"], 3)
        )
        assert steps[start + 3]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start + 3]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["diagnostic_cutoff_current"], 3)
        )

        # 1C discharge step for rpt cycle
        assert steps[start + 4]["StepType"] == "Dischrge"
        assert steps[start + 4]["StepMode"] == "Current"
        steps[start + 4]["StepValue"] = float(
            round(
                nominal_capacity
                * diagnostic_params["RPT_discharge_constant_current_2"],
                3,
            )
        )
        assert steps[start + 4]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[start + 4]["Ends"]["EndEntry"][0]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        # Third charge step for rpt cycle
        assert steps[start + 6]["StepType"] == "Charge"
        assert steps[start + 6]["StepMode"] == "Current"
        steps[start + 6]["StepValue"] = float(
            round(
                nominal_capacity * diagnostic_params["RPT_charge_constant_current"], 3
            )
        )
        steps[start + 6]["Limits"]["Voltage"] = float(
            round(diagnostic_params["diagnostic_charge_cutoff_voltage"], 3)
        )
        assert steps[start + 6]["Ends"]["EndEntry"][0]["EndType"] == "Current"
        steps[start + 6]["Ends"]["EndEntry"][0]["Value"] = float(
            round(nominal_capacity * diagnostic_params["diagnostic_cutoff_current"], 3)
        )

        # 2C discharge step for rpt cycle
        assert steps[start + 7]["StepType"] == "Dischrge"
        assert steps[start + 7]["StepMode"] == "Current"
        steps[start + 7]["StepValue"] = float(
            round(
                nominal_capacity
                * diagnostic_params["RPT_discharge_constant_current_3"],
                3,
            )
        )
        assert steps[start + 7]["Ends"]["EndEntry"][0]["EndType"] == "Voltage"
        steps[start + 7]["Ends"]["EndEntry"][0]["Value"] = float(
            round(diagnostic_params["diagnostic_discharge_cutoff_voltage"], 3)
        )

        return self

    def set_skip_to_end_diagnostic(self, max_v, min_v, step_key='070', new_step_key=None):
        """
        Helper function for setting the limits that cause the protocol to
        skip to the ending diagnostic.

        Args:
            max_v (float): Upper voltage limit to skip to ending diagnostic
            min_v (float): Lower voltage limit to skip to ending diagnostic
            step_key (str): Step for the ending diagnostic in order to
                recognize which EndEntry should be altered
            new_step_key (str): New value for goto step for the ending diagnostic

        Returns:
            dict
        """
        steps = self['MaccorTestProcedure']['ProcSteps']['TestStep']

        for step in steps:
            if step['Ends'] is not None and isinstance(step['Ends']['EndEntry'], list):
                for end in step['Ends']['EndEntry']:
                    if end['Step'] == step_key and end['EndType'] == 'Voltage' and end['Oper'] == '>=':
                        end['Value'] = max_v
                        if new_step_key:
                            end['Step'] = new_step_key
                    elif end['Step'] == step_key and end['EndType'] == 'Voltage' and end['Oper'] == '<=':
                        end['Value'] = min_v
                        if new_step_key:
                            end['Step'] = new_step_key

        return self


def insert_driving_parametersv1(reg_params, waveform_directory=None):
    """
    Args:
        reg_params (pandas.DataFrame): containing the following quantities
            discharge_profile (str): {'US06', 'LA4', '9Lap'}
            profile_charge_limit (float): upper limit voltage for the profile
            max_profile_power (float): maximum power setpoint during the profile
            n_repeats (int): number of repetitions for the profile
            discharge_cutoff_voltage (float): V
            power_scaling (float): Power relative to the other profiles
        waveform_directory (str): path to save waveform files
    """
    mwf_config = MACCOR_WAVEFORM_CONFIG
    velocity_name = reg_params["discharge_profile"] + "_velocity_waveform.txt"
    velocity_file = os.path.join(PROTOCOL_SCHEMA_DIR, velocity_name)
    df = convert_velocity_to_power_waveform(velocity_file, velocity_units="mph")

    if not os.path.exists(waveform_directory):
        os.makedirs(waveform_directory)

    mwf_config["value_scale"] = reg_params["max_profile_power"] * reg_params["power_scaling"]
    mwf_config["charge_limit_value"] = reg_params["profile_charge_limit"]
    mwf_config["charge_end_mode_value"] = reg_params["profile_charge_limit"] + 0.05
    mwf_config["discharge_end_mode_value"] = reg_params["discharge_cutoff_voltage"] - 0.1

    df = df[["time", "power"]]
    time_axis = list(df["time"]).copy()
    for i in range(reg_params['n_repeats'] - 1):
        time_axis = time_axis + [time_axis[-1] + el for el in df['time']]

    df = pd.DataFrame({'time': time_axis,
                       'power': list(df['power']) * reg_params['n_repeats']})
    filename = '{}_x{}_{}W'.format(reg_params['discharge_profile'], reg_params['n_repeats'],
                                   int(reg_params["max_profile_power"] * reg_params['power_scaling']))
    file_path = generate_maccor_waveform_file(df, filename, waveform_directory, mwf_config=mwf_config)

    return file_path


def insert_charging_parametersv1(reg_params, waveform_directory=None, max_c_rate=3.0, min_c_rate=0.2):
    """
    This function generates the waveform file for rapid charging. The charging rate parameters
    are applied over SOC windows defined by charge_start_soc, charge_soc_param_1,
    charge_soc_param_2, fast_charge_soc_limit.

    Args:
        reg_params (pandas.DataFrame): containing the following quantities
            charge_type_1 (str): {'smooth', 'step'} type of charging waveform
            charge_start_soc (float): assumed starting soc for the charge
            charge_current_param_1 (float): c-rate for first charging window
            charge_current_param_2 (float): c-rate for second charging window
            charge_current_param_3 (float): c-rate for third charging window
            charge_current_param_4 (float): c-rate for fourth charging window
            charge_soc_param_1 (float): soc point for dividing first charging window from second window
            charge_soc_param_2 (float): soc point for dividing second charging window from third window
            charge_fast_soc_limit (float): fraction of nominal capacity to end fast charging
            charge_cutoff_voltage (float): upper voltage limit for the charge
            capacity_nominal (float): expected capacity of cell for c-rate calculations
        waveform_directory (str): path to save waveform files
        max_c_rate (float): maximum charging c-rate for safety and cycler limits
        min_c_rate (float): minimum charging c-rate
    """
    mwf_config = {
        "control_mode": "I",
        "value_scale": 1,
        "charge_limit_mode": "V",
        "charge_limit_value": 4.2,
        "discharge_limit_mode": "V",
        "discharge_limit_value": 2.7,
        "charge_end_mode": "V",
        "charge_end_operation": ">=",
        "charge_end_mode_value": 4.25,
        "discharge_end_mode": "V",
        "discharge_end_operation": "<=",
        "discharge_end_mode_value": 2.5,
        "report_mode": "T",
        "report_value": 3.0000,
        "range": "A",
    }
    soc_initial = reg_params["charge_start_soc"]
    soc_final = reg_params["charge_fast_soc_limit"]
    charging_c_rates = [reg_params["charge_current_param_1"], reg_params["charge_current_param_2"],
                        reg_params["charge_current_param_3"], reg_params["charge_current_param_4"]]

    soc_points = [soc_initial, reg_params["charge_soc_param_1"], reg_params["charge_soc_param_2"], soc_final]
    final_c_rate = charging_c_rates[-1]

    charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)
    current_smooth, current_step, time_uniform = charging.get_currents_with_uniform_time_basis(charging_c_rates,
                                                                                               soc_points)

    assert np.max(current_smooth) <= max_c_rate, "Maximum c-rate exceeded in {}, abort".format(reg_params["seq_num"])
    assert np.max(current_step) <= max_c_rate, "Maximum c-rate exceeded in {}, abort".format(reg_params["seq_num"])

    if reg_params["charge_type_1"] == "smooth":
        df_charge = pd.DataFrame({"current": current_smooth, "time": time_uniform})
    elif reg_params["charge_type_1"] == "step":
        df_charge = pd.DataFrame({"current": current_step, "time": time_uniform})
    else:
        raise NotImplementedError

    if not os.path.exists(waveform_directory):
        os.makedirs(waveform_directory)

    mwf_config["value_scale"] = reg_params["capacity_nominal"] * max(df_charge["current"])
    mwf_config["charge_limit_value"] = reg_params["charge_cutoff_voltage"] - reg_params["charge_voltage_offset_1"]

    waveform_name = "{}_{}_{}".format(reg_params["project_name"], reg_params["charge_type_1"], reg_params["seq_num"])
    assert len(waveform_name) < 25, "Waveform name is too long"

    file_path = generate_maccor_waveform_file(
        df_charge,
        waveform_name,
        waveform_directory,
        mwf_config=mwf_config
    )

    return file_path


def generate_maccor_waveform_file(df, file_prefix, file_directory, mwf_config=None):
    """
    Helper function that takes in a variable power waveform and outputs a maccor waveform file (.MWF), which is read by
    the cycler when the procedure file has a "fast-wave" step.
    Relevant parameters to generate the .mwf files governing the input mode, charge/discharge limits, end conditions
    and scaling are stored in /conversion_schemas/maccor_waveform_conversion.yaml

    Args:
        df (pd.DataFrame): power waveform containing two columns "time" and "power", in sec and W respectively.
        file_prefix (str): prefix for the filename (extension is .MWF by default)
        file_directory (str): folder to store the mwf file
        mwf_config (dict): dictionary of instrument control parameters for generating the waveform

    Returns:
         path to the maccor waveform file generated
    """

    if mwf_config is None:
        mwf_config = MACCOR_WAVEFORM_CONFIG

    if "power" in df.columns:
        df["power"] = df["power"] / max(abs(df["power"])) * mwf_config["value_scale"]
        df["step_counter"] = df["power"].diff().fillna(0).ne(0).cumsum()
        df = df.groupby("step_counter").agg({"time": "count", "power": "first"})

        df.rename(columns={"time": "duration"}, inplace=True)

        df["control_mode"] = mwf_config["control_mode"]

        if mwf_config["control_mode"] == "P":
            df.loc[df["power"] == 0, "control_mode"] = "I"

        df["value"] = np.round(abs(df["power"]), 5)

        mask = df["power"] <= 0

    elif "current" in df.columns:
        df["current"] = df["current"] / max(abs(df["current"])) * mwf_config["value_scale"]
        df["step_counter"] = df["current"].diff().fillna(0).ne(0).cumsum()
        df = df.groupby("step_counter").agg({"time": "count", "current": "first"})

        df.rename(columns={"time": "duration"}, inplace=True)

        df["control_mode"] = mwf_config["control_mode"]

        df["value"] = np.round(abs(df["current"]), 5)

        mask = df["current"] <= 0

    df = df.assign(
        **{
            "state": "C",
            "limit_mode": mwf_config["charge_limit_mode"],
            "limit_value": mwf_config["charge_limit_value"],
            "end_mode": mwf_config["charge_end_mode"],
            "operation": mwf_config["charge_end_operation"],
            "end_mode_value": mwf_config["charge_end_mode_value"],
        }
    )

    df.loc[
        mask,
        [
            "state",
            "limit_mode",
            "limit_value",
            "end_mode",
            "operation",
            "end_mode_value",
        ],
    ] = [
        "D",
        mwf_config["discharge_limit_mode"],
        mwf_config["discharge_limit_value"],
        mwf_config["discharge_end_mode"],
        mwf_config["discharge_end_operation"],
        mwf_config["discharge_end_mode_value"],
    ]

    df["report_mode"] = mwf_config["report_mode"]
    df["report_value"] = mwf_config["report_value"]
    df["range"] = mwf_config["range"]

    MWF_file_path = os.path.join(file_directory, file_prefix + ".MWF")

    df[
        [
            "state",
            "control_mode",
            "value",
            "limit_mode",
            "limit_value",
            "duration",
            "end_mode",
            "operation",
            "end_mode_value",
            "report_mode",
            "report_value",
            "range",
        ]
    ].to_csv(MWF_file_path, sep="\t", header=None, index=None)

    return MWF_file_path



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Script for generating protocol files from
input parameters and procedure templates

Usage:
    generate_protocol [INPUT_JSON]

Options:
    -h --help        Show this screen
    --version        Show version


The `generate_protocol` script will generate a protocol file from input
parameters defined in the rows of a CSV-formatted input file.
It stores its outputs in `/data-share/protocols/`|
                                                |-`/procedures/`
                                                |-`/schedules/`
                                                |-`/names/`
For Maccor procedures the output procedures will be stored in `/data-share/protocols/procedures/`
For Arbin schedules the schedules will be stored in `/data-share/protocols/schedules/`
Additionally a file containing the names for the test files generated during the function call
will be stored in `/data-share/protocols/names/` with a date and time in the name to differentiate it
This file is to facilitate the process of starting tests on the cycling machines, by making it easier to
enter data into the appropriate fields.

The input json must contain the following fields
* `file_list` - filenames/paths corresponding to the input csv files

The output json will contain the following fields
* `file_list` - list of protocol files corresponding to the input files

Example:
$ generate_protocol '{"file_list": ["/data-share/raw/parameters/procedure_params_000112e3.csv"]}'
{
    "file_list": ["/data-share/protocols/procedures/name_1.000", "/data-share/protocols/procedures/name_2.000"]
}
"""

import os
import warnings
import datetime
import csv
import pandas as pd
from monty.os import makedirs_p
import xmltodict

from beep import logger
from beep.protocol import PROCEDURE_TEMPLATE_DIR, BIOLOGIC_TEMPLATE_DIR
from beep.protocol.maccor import Procedure, insert_driving_parametersv1, insert_charging_parametersv1
from beep.protocol.biologic import Settings


class ProtocolException(BaseException):
    """Errors related to generating protocol"""
    pass


def template_detection(filename, encoding="UTF-8"):
    """
    Reads a template and gathers metadata for use in decision logic

    Args:
        filename (str): Full path to template
        encoding (str): encoding used to read input file
    """
    if os.path.exists(filename):
        with open(filename, "rb") as f:
            text = f.read().decode(encoding)
        template_data = xmltodict.parse(text, process_namespaces=False, strip_whitespace=True)
        template_length = len(template_data["MaccorTestProcedure"]["ProcSteps"]["TestStep"])
    else:
        template_length = None
    return template_length


def generate_protocol_files_from_csv(csv_filename, output_directory=None):

    """
    Generates a set of protocol files from csv filename input by
    reading protocol file input corresponding to each line of
    the csv file. Writes a csv file that.

    Args:
        csv_filename (str): CSV containing protocol file parameters.
        output_directory (str): directory in which to place the output files
    """
    # Read csv file
    protocol_params_df = pd.read_csv(csv_filename)

    successfully_generated_files = []
    file_generation_failures = []
    names = []
    result = ""
    message = {"comment": "", "error": ""}
    if output_directory is None:
        output_directory = PROCEDURE_TEMPLATE_DIR

    # Create required directories if not already present
    for subdir in ["settings", "procedures", "names"]:
        subdir_path = os.path.abspath(os.path.join(output_directory, subdir))
        if not os.path.exists(subdir_path):
            makedirs_p(subdir_path)

    for index, protocol_params in protocol_params_df.iterrows():
        template = protocol_params["template"]
        protocol = None
        # Filename for the output
        filename_prefix = "_".join(
            [
                protocol_params["project_name"],
                "{:06d}".format(protocol_params["seq_num"]),
            ]
        )
        if ".000" in template:  # Extension for maccor procedure files
            template_fullpath = os.path.join(PROCEDURE_TEMPLATE_DIR, template)
            template_length = template_detection(template_fullpath)
            if "diagnostic_parameter_set" in protocol_params:  # For parameters include diagnostics load those values
                diag_params_df = pd.read_csv(
                    os.path.join(PROCEDURE_TEMPLATE_DIR, "PreDiag_parameters - DP.csv")
                )
                diagnostic_params = diag_params_df[
                    diag_params_df["diagnostic_parameter_set"]
                    == protocol_params["diagnostic_parameter_set"]
                    ].squeeze()

            if template_length == 23 and template == "EXP.000":  # length and name for initial procedure files
                protocol = Procedure.from_exp(
                    **protocol_params[["cutoff_voltage", "charge_rate", "discharge_rate"]]
                )
            elif template_length == 72:  # length for V1 and V1 diagnostic templates without ending diagnostics
                protocol = Procedure.from_regcyclev2(protocol_params)
                protocol.add_procedure_diagcyclev2(
                    protocol_params["capacity_nominal"], diagnostic_params
                )
            elif template_length == 96:  # template length for diagnostic type cycling
                mwf_dir = os.path.join(output_directory, "mwf_files")
                if protocol_params["project_name"] == "RapidC":  # Project with charging waveform
                    waveform_name = insert_charging_parametersv1(protocol_params,
                                                                 waveform_directory=mwf_dir)
                    protocol = Procedure.generate_procedure_chargingv1(index,
                                                                       protocol_params,
                                                                       waveform_name,
                                                                       template=template_fullpath)
                elif protocol_params["project_name"] == "Drive":  # Project with discharging waveform
                    waveform_name = insert_driving_parametersv1(protocol_params,
                                                                waveform_directory=mwf_dir)
                    protocol = Procedure.generate_procedure_drivingv1(index,
                                                                      protocol_params,
                                                                      waveform_name,
                                                                      template=template_fullpath)
                else:  # Use the default parameterization for PreDiag/Prediction Diagnostic projects
                    protocol = Procedure.generate_procedure_regcyclev3(index,
                                                                       protocol_params,
                                                                       template=template_fullpath)
                protocol.generate_procedure_diagcyclev3(
                    protocol_params["capacity_nominal"], diagnostic_params
                )
            else:  # Case where its not possible to match the procedure template
                failure = {
                    "comment": "Unable to find template: " + template,
                    "error": "Not Found",
                }
                file_generation_failures.append(failure)
                warnings.warn("Unsupported file template {}, skipping.".format(template))
                result = "error"
                continue

            filename = "{}.000".format(filename_prefix)
            filename = os.path.join(output_directory, "procedures", filename)

        elif ".mps" in template and template == "formationV1.mps":  # biologic settings template and formation project
            protocol = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, template))
            protocol = protocol.formation_protocol_bcs(protocol_params)
            filename = "{}.mps".format(filename_prefix)
            filename = os.path.join(output_directory, "settings", filename)
        elif ".sdu" in template:  # No schedule file templates implemented
            failure = {
                "comment": "Schedule file generation is not yet implemented",
                "error": "Not Implemented"
            }
            file_generation_failures.append(failure)
            logger.warning("Schedule file generation not yet implemented")
            result = "error"
            continue
        else:  # Unable to match to any known template format
            failure = {
                "comment": "Unable to find template: " + template,
                "error": "Not Found",
            }
            file_generation_failures.append(failure)
            warnings.warn("Unsupported file template {}, skipping.".format(template))
            result = "error"
            continue

        logger.info(filename)
        protocol.to_file(filename)
        successfully_generated_files.append(filename)
        names.append(filename_prefix + "_")

    # This block of code produces the file containing all of the run file
    # names produced in this function call. This is to make starting tests easier
    _, namefile = os.path.split(csv_filename)
    namefile = namefile.split("_")[0] + "_names_"
    namefile = namefile + datetime.datetime.now().strftime("%Y%m%d_%H%M") + ".csv"

    names_dir = os.path.join(output_directory, "names")
    os.makedirs(names_dir, exist_ok=True)

    with open(os.path.join(names_dir, namefile), "w", newline="") as outputfile:
        wr = csv.writer(outputfile)
        for name in names:
            wr.writerow([name])
    outputfile.close()

    num_generated_files = len(successfully_generated_files)
    num_generation_failures = len(file_generation_failures)
    num_files = num_generated_files + num_generation_failures

    message = {
        "comment": "Generated {} of {} protocols".format(num_generated_files, num_files),
        "error": ""
    }
    if not result:
        result = "success"
    else:
        message["error"] = "Failed to generate {} of {} protocols".format(num_generation_failures, num_files)
        logger.error(message["error"])

    return successfully_generated_files, file_generation_failures, result, message



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to the fitting functions that extract cell parameters from dis profiles"""

import os
import unittest
import numpy as np
import pandas as pd
from scipy.optimize import differential_evolution
from beep.structure.cli import auto_load_processed
from beep.features.intracell_analysis import IntracellAnalysis, \
    get_halfcell_voltages
from beep.features.intracell_losses import IntracellCycles, IntracellFeatures

from beep.tests.constants import TEST_FILE_DIR


class IntracellAnalysisTest(unittest.TestCase):
    def setUp(self):
        run = os.path.join(TEST_FILE_DIR,
                           'PreDiag_000220_00005E_structure_omit.json')
        self.cell_struct = auto_load_processed(run)
        self.cathode_file = os.path.join(TEST_FILE_DIR,
                                         'data-share/raw/cell_info/cathode_test.csv')
        self.anode_file = os.path.join(TEST_FILE_DIR,
                                       'data-share/raw/cell_info/anode_test.csv')

    def tearDown(self):
        pass

    @unittest.skip
    def test_process_beep_cycle_data_for_initial_halfcell_analysis_mock(self):
        ia = IntracellAnalysis(
            self.cathode_file,
            self.anode_file,
            cycle_type='rpt_0.2C',
            step_type=0
        )
        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            ia.process_beep_cycle_data_for_initial_halfcell_analysis(
                self.cell_struct, step_type=0)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile_aligned['Voltage_aligned'].min(),
            2.742084, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile_aligned['Voltage_aligned'].max(),
            4.196994, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['Voltage'].min(), 2.703006, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['Voltage'].max(), 4.196994, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['charge_capacity'].min(), 0.0, 4)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['charge_capacity'].max(), 4.539547,
            5)

    def test_intracell_halfcell_matching_v2_mock(self):
        ia = IntracellAnalysis(
            self.cathode_file,
            self.anode_file,
            cycle_type='rpt_0.2C',
            step_type=0
        )
        #
        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            ia.process_beep_cycle_data_for_initial_halfcell_analysis(
                self.cell_struct)

        test_opt = np.array([0.999459, -4.1740795, 1.0, 0.1, 0.1])
        (PE_pristine_matched,
         NE_pristine_matched,
         df_real_interped,
         emulated_full_cell_interped) = ia.halfcell_initial_matching_v2(
            test_opt,
            real_cell_initial_charge_profile_aligned,
            ia.pe_pristine,
            ia.ne_1_pristine,
            ia.ne_2_pristine_pos,
            ia.ne_2_pristine_neg)

        self.assertAlmostEqual(PE_pristine_matched['SOC_aligned'].min(),
                               -93.08943089430896, 5)
        self.assertAlmostEqual(PE_pristine_matched['SOC_aligned'].max(),
                               110.16260162601628, 5)
        self.assertAlmostEqual(PE_pristine_matched['Voltage_aligned'].min(),
                               2.865916, 5)
        self.assertAlmostEqual(PE_pristine_matched['Voltage_aligned'].max(),
                               4.299219386998656, 5)

    def test_intracell_get_dq_dv_mock(self):
        ia = IntracellAnalysis(
            self.cathode_file,
            self.anode_file,
            cycle_type='rpt_0.2C',
            step_type=0
        )
        #
        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            ia.process_beep_cycle_data_for_initial_halfcell_analysis(
                self.cell_struct)

        test_opt = np.array([0.999459, -4.1740795, 1.0, 0.1, 0.1])
        (PE_p_m, NE_p_m, df_real_i,
         em_interped) = ia.halfcell_initial_matching_v2(
            test_opt,
            real_cell_initial_charge_profile_aligned,
            ia.pe_pristine,
            ia.ne_1_pristine,
            ia.ne_2_pristine_pos,
            ia.ne_2_pristine_neg)

        real_cell_candidate_charge_profile_aligned = ia.process_beep_cycle_data_for_candidate_halfcell_analysis(
            self.cell_struct,
            real_cell_initial_charge_profile_aligned,
            real_cell_initial_charge_profile,
            3)

        self.assertAlmostEqual(
            real_cell_candidate_charge_profile_aligned["Voltage_aligned"].min(),
            2.742084, 5)

        (PE_out_centered,
         NE_out_centered,
         dVdQ_over_SOC_real,
         dVdQ_over_SOC_emulated,
         df_real_interped,
         emulated_full_cell_interped) = ia.get_dQdV_over_V_from_degradation_matching(
            test_opt, PE_p_m, NE_p_m,
            ia.ne_2_pristine_pos,
            ia.ne_2_pristine_neg,
            real_cell_candidate_charge_profile_aligned)
        print(PE_out_centered)
        self.assertAlmostEqual(PE_out_centered["Voltage_aligned"].min(),
                               2.86591646, 5)
        self.assertAlmostEqual(NE_out_centered["Voltage_aligned"].min(),
                               0.011226739, 5)
        self.assertAlmostEqual(dVdQ_over_SOC_real["Voltage_aligned"].mean(),
                               3.45, 5)
        self.assertAlmostEqual(dVdQ_over_SOC_emulated["Voltage_aligned"].min(),
                               2.7, 5)
        self.assertAlmostEqual(dVdQ_over_SOC_emulated["Voltage_aligned"].max(),
                               4.2, 5)
        #
        (
            PE_upper_voltage, PE_lower_voltage, PE_upper_SOC, PE_lower_SOC,
            PE_mass,
            NE_upper_voltage, NE_lower_voltage, NE_upper_SOC, NE_lower_SOC,
            NE_mass,
            SOC_upper, SOC_lower, Li_mass) = get_halfcell_voltages(
            PE_out_centered,
            NE_out_centered)
        self.assertAlmostEqual(PE_upper_voltage, 4.28427357, 5)
        self.assertAlmostEqual(PE_lower_voltage, 3.56538024, 5)
        self.assertAlmostEqual(PE_upper_SOC, 99.0605427, 5)
        self.assertAlmostEqual(PE_lower_SOC, 48.851774, 5)
        self.assertAlmostEqual(PE_mass, 202.843024, 5)
        self.assertAlmostEqual(NE_upper_voltage, 0.0828635, 5)
        self.assertAlmostEqual(NE_lower_voltage, 0.86598357, 5)
        self.assertAlmostEqual(NE_upper_SOC, 95.1648351, 5)
        self.assertAlmostEqual(NE_lower_SOC, 42.30769230, 5)
        self.assertAlmostEqual(NE_mass, 192.6796998, 5)
        self.assertAlmostEqual(SOC_upper, 183.363318, 5)
        self.assertAlmostEqual(SOC_lower, 81.518334, 5)
        self.assertAlmostEqual(PE_upper_SOC, 99.06054279, 5)
        self.assertAlmostEqual(PE_lower_SOC, 48.8517745, 5)
        self.assertAlmostEqual(Li_mass, 185.2689422, 5)

    def test_process_beep_cycle_data_for_initial_halfcell_analysis(self):
        ia = IntracellAnalysis(os.path.join(TEST_FILE_DIR,
                                            'cathode_clean_cc_charge_exptl_aligned.csv'),
                               os.path.join(TEST_FILE_DIR,
                                            'anode_secondMeasure_clean_cc_charge_exptl_aligned.csv'),
                               cycle_type='rpt_0.2C',
                               step_type=0
                               )
        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            ia.process_beep_cycle_data_for_initial_halfcell_analysis(
                self.cell_struct, step_type=0)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile_aligned['Voltage_aligned'].min(),
            2.742084, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile_aligned['Voltage_aligned'].max(),
            4.196994, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['Voltage'].min(), 2.703006, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['Voltage'].max(), 4.196994, 5)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['charge_capacity'].min(), 0.0, 4)
        self.assertAlmostEqual(
            real_cell_initial_charge_profile['charge_capacity'].max(), 4.539547,
            5)

    def test_intracell(self):
        ia = IntracellAnalysis(os.path.join(TEST_FILE_DIR,
                                            'cathode_clean_cc_charge_exptl_aligned.csv'),
                               os.path.join(TEST_FILE_DIR,
                                            'anode_secondMeasure_clean_cc_charge_exptl_aligned.csv'),
                               cycle_type='rpt_0.2C',
                               step_type=0
                               )
        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            ia.process_beep_cycle_data_for_initial_halfcell_analysis(
                self.cell_struct)

        # Solving initial electrode matching to real full cell
        electrode_matching_bounds = (
            (0.8, 1.2), (-20.0, 20.0), (1, 1), (0.1, 0.1), (0.1, 0.1))
        # (-1,1),(0.01,0.1),(0.01,0.1)
        opt_result_halfcell_initial_matching = differential_evolution(
            ia._get_error_from_halfcell_initial_matching,
            electrode_matching_bounds,
            args=(real_cell_initial_charge_profile_aligned,
                  ia.pe_pristine, ia.ne_1_pristine,
                  ia.ne_2_pristine_pos, ia.ne_2_pristine_neg),
            strategy='best1bin', maxiter=1000,
            popsize=15, tol=0.001, mutation=0.5,
            recombination=0.7, seed=1,
            callback=None, disp=False, polish=True,
            init='latinhypercube', atol=0,
            updating='deferred', workers=-1, constraints=())
        print(opt_result_halfcell_initial_matching)

        self.assertEqual(opt_result_halfcell_initial_matching.success, True)
        self.assertAlmostEqual(opt_result_halfcell_initial_matching.x[0],
                               0.999459, 5)
        self.assertAlmostEqual(opt_result_halfcell_initial_matching.x[1],
                               -4.1740795, 6)

        # test_opt = np.array([0.999459, -4.1740795, 1.0, 0.1, 0.1])
        (PE_pristine_matched,
         NE_pristine_matched,
         df_real_interped,
         emulated_full_cell_interped) = ia.halfcell_initial_matching_v2(
            opt_result_halfcell_initial_matching.x,
            real_cell_initial_charge_profile_aligned,
            ia.pe_pristine,
            ia.ne_1_pristine,
            ia.ne_2_pristine_pos,
            ia.ne_2_pristine_neg)
        print(PE_pristine_matched)
        self.assertAlmostEqual(PE_pristine_matched['SOC_aligned'].min(),
                               -4.675029, 5)
        self.assertAlmostEqual(PE_pristine_matched['SOC_aligned'].max(),
                               109.350057, 5)
        self.assertAlmostEqual(PE_pristine_matched['Voltage_aligned'].min(),
                               2.865916, 5)
        self.assertAlmostEqual(PE_pristine_matched['Voltage_aligned'].max(),
                               4.29917115, 5)

        eol_cycle_index_list = self.cell_struct.diagnostic_summary[
            (self.cell_struct.diagnostic_summary.cycle_type == ia.cycle_type) &
            (
                    self.cell_struct.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()
        #
        # initial bounds (for first cycle); they expand as the cell degrades (see update further down).
        # allow negative degradation, as it may reflect gained capacity from electrolyte wetting or other phenomena
        degradation_bounds = ((-10, 50),  # LLI
                              (-10, 50),  # LAM_PE
                              (-10, 50),  # LAM_NE
                              (1, 1),  # (-1,1) x_NE_2
                              (0.1, 0.1),  # (0.01,0.1)
                              (0.1, 0.1),  # (0.01,0.1)
                              )
        #
        # # initializations before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list):
            real_cell_candidate_charge_profile_aligned = ia.process_beep_cycle_data_for_candidate_halfcell_analysis(
                self.cell_struct,
                real_cell_initial_charge_profile_aligned,
                real_cell_initial_charge_profile,
                cycle_index)

            degradation_optimization_result = differential_evolution(
                ia._get_error_from_degradation_matching,
                degradation_bounds,
                args=(PE_pristine_matched,
                      NE_pristine_matched,
                      ia.ne_2_pristine_pos,
                      ia.ne_2_pristine_neg,
                      real_cell_candidate_charge_profile_aligned
                      ),
                strategy='best1bin', maxiter=100000,
                popsize=15, tol=0.001, mutation=0.5,
                recombination=0.7,
                seed=1,
                callback=None, disp=False, polish=True,
                init='latinhypercube',
                atol=0, updating='deferred', workers=-1,
                constraints=()
            )

            (PE_out_centered,
             NE_out_centered,
             dVdQ_over_SOC_real,
             dVdQ_over_SOC_emulated,
             df_real_interped,
             emulated_full_cell_interped) = ia.get_dQdV_over_V_from_degradation_matching(
                degradation_optimization_result.x,
                PE_pristine_matched,
                NE_pristine_matched,
                ia.ne_2_pristine_pos,
                ia.ne_2_pristine_neg,
                real_cell_candidate_charge_profile_aligned)
            #
            (PE_upper_voltage, PE_lower_voltage, PE_upper_SOC, PE_lower_SOC,
             PE_mass,
             NE_upper_voltage, NE_lower_voltage, NE_upper_SOC, NE_lower_SOC,
             NE_mass,
             SOC_upper, SOC_lower, Li_mass) = get_halfcell_voltages(
                PE_out_centered, NE_out_centered)
            #
            LLI = degradation_optimization_result.x[0]
            LAM_PE = degradation_optimization_result.x[1]
            LAM_NE = degradation_optimization_result.x[2]
            x_NE_2 = degradation_optimization_result.x[3]
            alpha_real = degradation_optimization_result.x[4]
            alpha_emulated = degradation_optimization_result.x[5]

            tmp_dict = {cycle_index: [LLI, LAM_PE, LAM_NE, x_NE_2, alpha_real,
                                      alpha_emulated,
                                      PE_upper_voltage, PE_lower_voltage,
                                      PE_upper_SOC, PE_lower_SOC, PE_mass,
                                      NE_upper_voltage, NE_lower_voltage,
                                      NE_upper_SOC, NE_lower_SOC, NE_mass,
                                      Li_mass
                                      ]
                        }

            dataset_dict_of_cell_degradation_path.update(tmp_dict)
            real_cell_dict_of_profiles.update(
                {cycle_index: real_cell_candidate_charge_profile_aligned})

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['LLI', 'LAM_PE', 'LAM_NE',
                                             'x_NE_2', 'alpha_real',
                                             'alpha_emulated',
                                             'PE_upper_voltage',
                                             'PE_lower_voltage', 'PE_upper_SOC',
                                             'PE_lower_SOC',
                                             'PE_mass', 'NE_upper_voltage',
                                             'NE_lower_voltage', 'NE_upper_SOC',
                                             'NE_lower_SOC', 'NE_mass',
                                             'Li_mass'
                                             ]).T
        # print(degradation_df.iloc[0].to_list())
        self.assertAlmostEqual(degradation_df['LLI'].iloc[0], -0.027076, 5)
        self.assertAlmostEqual(degradation_df['LAM_PE'].iloc[0], 0.06750165, 5)
        self.assertAlmostEqual(degradation_df['LAM_NE'].iloc[0], 0.3055425, 5)
        self.assertAlmostEqual(degradation_df['PE_upper_voltage'].iloc[0],
                               4.25095116, 5)
        self.assertAlmostEqual(degradation_df['PE_lower_voltage'].iloc[0],
                               3.62354913951, 5)
        self.assertAlmostEqual(degradation_df['PE_upper_SOC'].iloc[0],
                               95.7202505, 5)
        self.assertAlmostEqual(degradation_df['PE_mass'].iloc[0], 109.24224079,
                               5)

        self.assertAlmostEqual(degradation_df['NE_upper_voltage'].iloc[0],
                               0.050515360, 5)
        self.assertAlmostEqual(degradation_df['NE_lower_voltage'].iloc[0],
                               0.8564127166, 5)
        self.assertAlmostEqual(degradation_df['NE_upper_SOC'].iloc[0],
                               91.832460, 5)
        self.assertAlmostEqual(degradation_df['NE_mass'].iloc[0], 108.900146, 5)

        self.assertAlmostEqual(degradation_df['Li_mass'].iloc[0], 104.680978, 5)

    def test_intracell_wrappers(self):
        ia = IntracellAnalysis(
            os.path.join(TEST_FILE_DIR, 'data-share', 'raw', 'cell_info',
                         'cathode_test.csv'),
            os.path.join(TEST_FILE_DIR, 'data-share', 'raw', 'cell_info',
                         'anode_test.csv'),
            cycle_type='rpt_0.2C',
            step_type=0
        )

        (cell_init_aligned, cell_init_profile, PE_matched,
         NE_matched) = ia.intracell_wrapper_init(self.cell_struct)

        eol_cycle_index_list = self.cell_struct.diagnostic_summary[
            (self.cell_struct.diagnostic_summary.cycle_type == ia.cycle_type) &
            (
                    self.cell_struct.diagnostic_summary.discharge_capacity > ia.THRESHOLD)
            ].cycle_index.to_list()

        # # initializations before for loop
        dataset_dict_of_cell_degradation_path = dict()
        real_cell_dict_of_profiles = dict()
        for i, cycle_index in enumerate(eol_cycle_index_list):
            loss_dict, profiles_dict = ia.intracell_values_wrapper(cycle_index,
                                                                   self.cell_struct,
                                                                   cell_init_aligned,
                                                                   cell_init_profile,
                                                                   PE_matched,
                                                                   NE_matched,
                                                                   )
            dataset_dict_of_cell_degradation_path.update(loss_dict)
            real_cell_dict_of_profiles.update(profiles_dict)

        degradation_df = pd.DataFrame(dataset_dict_of_cell_degradation_path,
                                      index=['LLI', 'LAM_PE', 'LAM_NE',
                                             'x_NE_2', 'alpha_real',
                                             'alpha_emulated',
                                             'PE_upper_voltage',
                                             'PE_lower_voltage', 'PE_upper_SOC',
                                             'PE_lower_SOC',
                                             'PE_mass', 'NE_upper_voltage',
                                             'NE_lower_voltage', 'NE_upper_SOC',
                                             'NE_lower_SOC', 'NE_mass',
                                             'Li_mass'
                                             ]).T
        print(degradation_df['LLI'])
        print(degradation_df['LAM_PE'])
        print(degradation_df['Li_mass'])
        self.assertAlmostEqual(degradation_df['LLI'].iloc[0], -9.999983, 5)
        self.assertAlmostEqual(degradation_df['LLI'].iloc[1], -9.999556, 5)
        self.assertAlmostEqual(degradation_df['LAM_PE'].iloc[0], 49.984768, 5)
        self.assertAlmostEqual(degradation_df['LAM_PE'].iloc[1], 49.984877, 5)
        self.assertAlmostEqual(degradation_df["Li_mass"].iloc[1], 12.312480, 3)

        # Values for real anode and cathode measurements
        # self.assertAlmostEqual(degradation_df['LLI'].iloc[0], -0.027076, 5)
        # self.assertAlmostEqual(degradation_df['LLI'].iloc[1], 2.712016, 5)
        # self.assertAlmostEqual(degradation_df['LAM_PE'].iloc[0], 0.06750165, 5)
        # self.assertAlmostEqual(degradation_df['LAM_PE'].iloc[1], 2.745720, 5)
        # self.assertAlmostEqual(degradation_df["Li_mass"].iloc[1], 101.82022, 3)


class IntracellFeaturesTest(unittest.TestCase):
    def setUp(self):
        run_path = os.path.join(TEST_FILE_DIR,
                                'PreDiag_000220_00005E_structure_omit.json')
        self.datapath = auto_load_processed(run_path)
        cathode_file = os.path.join(TEST_FILE_DIR,
                                    'data-share/raw/cell_info/cathode_test.csv')
        anode_file = os.path.join(TEST_FILE_DIR,
                                  'data-share/raw/cell_info/anode_test.csv')
        self.params = {
            'diagnostic_cycle_type': 'rpt_0.2C',
            'step_type': 0,
            "anode_file": anode_file,
            "cathode_file": cathode_file
        }

    def test_IntracellCycles(self):
        featurizer = IntracellCycles(self.datapath, self.params)
        featurizer.create_features()
        X = featurizer.features
        self.assertEqual(X.shape, (2, 17))
        self.assertAlmostEqual(X["LLI"].iloc[0], -9.999983, 5)
        self.assertAlmostEqual(X["Li_mass"].iloc[1], 12.312480, 3)

    def test_IntracellFeatures(self):
        featurizer = IntracellFeatures(self.datapath, self.params)
        featurizer.create_features()
        X = featurizer.features
        self.assertEqual(X.shape, (1, 34))
        self.assertAlmostEqual(X["diag_0_LLI"].iloc[0], -9.999983, 5)
        self.assertAlmostEqual(X["diag_1_Li_mass"].iloc[0], 12.312480, 3)

    def test_validation(self):
        # Modify datapath_run to be invalid
        mask = self.datapath.diagnostic_summary.cycle_type == "rpt_0.2C"
        self.datapath.diagnostic_summary.loc[mask, "discharge_capacity"] = 3.35

        featurizer = IntracellFeatures(self.datapath, self.params)
        val, msg = featurizer.validate()
        self.assertFalse(val)



"""
For running ML experiments on pregenerated feature files and
reading/writing static files for battery prediction given a feature set.
"""
import copy
import pprint
from typing import List, Union, Iterable, Tuple
from math import sqrt

import numpy as np
import pandas as pd
from monty.json import MSONable
from monty.serialization import loadfn, dumpfn
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import (
    Lasso,
    LassoCV,
    RidgeCV,
    Ridge,
    ElasticNetCV,
    ElasticNet,
    MultiTaskElasticNet,
    MultiTaskElasticNetCV,
)
from sklearn.model_selection import train_test_split
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    max_error,
    r2_score
)

from beep import logger
from beep.features.base import BEEPFeatureMatrix


class BEEPMLExperimentError(BaseException):
    """Raise when error arises specific to BEEP ML Experiments"""
    pass


class BEEPLinearModelExperiment(MSONable):
    """A class for training, predicting, managing, and (de)serializing
    BEEP-based linear ML battery cycler experiments.


    1 BEEPMLExperiment = 1 model on 1 set of training data
        - Can predict on any number of prediction sets

    Args:
        feature_matrix (BEEPFeatureMatrix): The feature matrix for learning.
        target_matrix (BEEPFeatureMatrix): A matrix of targets for learning.
        targets ([str]): List of string target names. Should be found in the
            target_matrix dataframe.
        model_name (str): String specifying the linear model to use.
        alphas (str): Alpha values to try durig hyperparameter optimization.
        train_feature_drop_nan_thresh (float): Threshold 0-1 fraction of
            samples that must be present for a feature in order to avoid
            being dropped. Applies to training set only.
        train_sample_drop_nan_thresh (float): Threshold 0-1 fraction of
            features that must be present in order for a sample to avoid
            being dropped. Applies to training data only.
        predict_sample_nan_thresh (float): Threshold 0-1 fraction of
            features that must be present in order for a sample to avoid
            being dropped. Applies to prediction data only.
        drop_nan_training_targets (float): Drop nan training targets samples.
        impute_strategy (str): Define the strategy for imputing unknown
            or nan values in the feature matrix. Applies to both training
            and prediction.
        kfold (int): Number of folds k to use in running hyperparameter
            optimization with cross validation.
        max_iter (int): Number of iterations to use in determining
            optimal hyperparameters during cross validation.
        tol (float): Tolerance for CV-based hyperparameter optimization.
        l1_ratio ([float]): Ratios of L1/L2 losses to explore during hyperparameter
            optimization.
        homogenize_features (bool): Allow features generated with mismatching
            hyperparameters to be coalesced into the same "feature". I.e.,
            features generated with SomeFeaturizer(param=1) to be used as
            features generated with SomeFeaturizer(param=2).
    """
    ALLOWED_MODELS = ("elasticnet", "ridge", "lasso")
    ERROR_METRICS = {
        "rmse": lambda x: sqrt(mean_squared_error(*x)),
        "mae": lambda x: mean_absolute_error(*x),
        "r2": lambda x: r2_score(*x),
        "max_error": lambda x: max_error(*x)
    }

    def __init__(
            self,
            feature_matrix: BEEPFeatureMatrix,
            target_matrix: BEEPFeatureMatrix,
            targets: List[str],
            model_name: str,
            alphas: Union[None, Iterable[float]] = None,
            train_feature_drop_nan_thresh: float = 0.95,
            train_sample_drop_nan_thresh: float = 0.50,
            predict_sample_nan_thresh: float = 0.75,
            drop_nan_training_targets: bool = False,
            impute_strategy: str = "median",
            kfold: int = 5,
            max_iter: int = 1e6,
            tol: float = 1e-4,
            # only relevant for elasticnet
            l1_ratio: Union[Tuple[float], List[float]] = (
            0.001, 0.1, 0.5, 0.7, 0.9, 0.95, 1),
            homogenize_features: bool = True
    ):
        if model_name not in self.ALLOWED_MODELS:
            raise ValueError(
                f"Model {model_name} not supported by {self.__class__.__name__}")

        if len(targets) < 1:
            raise ValueError("At least one target must be specified")

        self.feature_matrix = feature_matrix
        self.target_matrix = target_matrix

        X = self.feature_matrix.matrix.replace([np.inf, -np.inf], np.nan)
        y = self.target_matrix.matrix.replace([np.inf, -np.inf], np.nan)

        if homogenize_features:
            X = self._remove_param_hash_from_features(X)
            y = self._remove_param_hash_from_features(y)

        if X.shape[0] != y.shape[0]:
            raise BEEPMLExperimentError(
                "Can't run experiment on unequal numbers of input samples."
            )
        if X.shape[0] < X.shape[1]:
            logger.warning(
                f"Number of samples ({X.shape[0]}) less than number of "
                f"features ({X.shape[1]}); may cause overfitting."
            )

        # Form the clean feature matrix
        X = X.dropna(axis=1, thresh=train_sample_drop_nan_thresh * X.shape[0])
        X = X.dropna(axis=0, thresh=train_sample_drop_nan_thresh * X.shape[1])
        X = self._impute_df(X, method=impute_strategy)
        self.impute_strategy = impute_strategy
        if X.shape[0] < 2 or X.shape[1] < 1:
            raise BEEPMLExperimentError(
                "Cleaned feature matrix has dimensions of less "
                "than 1 feature or less than 2 samples. Try adjusting "
                "the thresholds for cleaning or examine your feature "
                "matrix."
            )

        # Form the clean target matrix
        missing_targets = [t for t in targets if t not in y.columns]
        if missing_targets:
            raise BEEPMLExperimentError(
                f"Required target columns missing from "
                f"target matrix: {missing_targets}"
            )
        y = y[targets].loc[X.index]
        if y.isna().any().any():
            if drop_nan_training_targets:
                y = y.dropna(axis=0)
            else:
                raise BEEPMLExperimentError(
                    "Target matrix contains nans and drop_nan_targets is "
                    "set to False."
                )
        if y.shape[0] < 2:
            raise BEEPMLExperimentError(
                "Target matrix after dropping nans is less than 2 samples."
            )

        # Ensure there will be an equal number of X samples
        # and y samples
        self.X = X.loc[y.index]
        self.y = y

        # These features must be present in passed dfs for predictions to work
        self.feature_labels = self.X.columns.tolist()

        self.targets = targets

        self.multi = len(self.targets) > 1

        if self.multi and model_name != "elasticnet":
            raise BEEPMLExperimentError(
                f"Model {model_name} not supported for multiple target "
                f"regression."
            )

        self.model_name = model_name if model_name else "elasticnet"
        self.model = None

        self.train_feature_drop_thresh = train_feature_drop_nan_thresh
        self.train_sample_drop_thresh = train_sample_drop_nan_thresh
        self.predict_sample_nan_thresh = predict_sample_nan_thresh
        self.drop_nan_training_targets = drop_nan_training_targets

        # todo: this is only to help with deserialization, this could cause
        # todo: contamination in judging test scores when used with
        # todo: train_and_score()
        self.scaler = StandardScaler().fit(X)
        self.kfold = kfold
        self.alphas = alphas
        self.max_iter = max_iter
        self.tol = tol
        self.l1_ratio = l1_ratio

        self.optimal_hyperparameters = None
        self.homogenize_features = homogenize_features

    def train(self, X: pd.DataFrame = None, y: pd.DataFrame = None):
        """Train on 100% of available data.

        Args:
            X (pd.Dataframe): Clean and homogenized learning features.
                If not specified, df defined in __init__ (all training
                data) is used.
            y (pd.DataFrame): Clean and homogenized targets. If not
                specified, df defined in __init__ (all training data)
                is used.

        Returns:
            model (BaseEstimator): The sklearn model, fit on training data.
            training_errors (dict): Training errors based on multiple metrics.

        """
        X = X if X is not None else self.X
        y = y if y is not None else self.y

        if not self.multi:
            y = y[self.targets[0]]

        X = self.scaler.fit_transform(X)

        logger.info(
            f"Training on {X.shape[0]} samples with {X.shape[1]} features "
            f"predicting {y.shape[0]}"
        )

        kwargs = {
            "fit_intercept": True,
            "alphas": self.alphas,
            "cv": self.kfold,
            "max_iter": self.max_iter,
            "tol": self.tol,
            "l1_ratio": self.l1_ratio
        }

        if self.model_name == "elasticnet":
            if self.multi:
                cv_class = MultiTaskElasticNetCV
                model_class = MultiTaskElasticNet
            else:
                cv_class = ElasticNetCV
                model_class = ElasticNet
        elif self.model_name == "lasso":
            cv_class = LassoCV
            model_class = Lasso
            kwargs.pop("l1_ratio")
        elif self.model_name == "ridge":
            cv_class = RidgeCV
            model_class = Ridge
            kwargs.pop("l1_ratio")
            kwargs.pop("max_iter")
            kwargs.pop("tol")

            # Ridge has to have alphas set by hand as it has no
            # default alphas
            if not kwargs["alphas"]:
                kwargs["alphas"] = (1e-3, 1e-2, 1e-1, 1, 10, 100, 1000)
        else:
            raise NotImplementedError(f"Unsupported model '{self.model_name}'")

        # Search for optimal hyperparameters
        cv = cv_class(**kwargs)
        cv.fit(X, y)

        # Set optimal hyperparameters and refit
        optimal_hyperparameters = {"alpha": cv.alpha_}
        if self.model_name == "elasticnet":
            optimal_hyperparameters["l1_ratio"] = cv.l1_ratio_

        model_kwargs = {
            "fit_intercept": True,
            "normalize": False,
            "max_iter": self.max_iter,
        }
        model_kwargs.update(optimal_hyperparameters)
        self.optimal_hyperparameters = optimal_hyperparameters

        model = model_class(**model_kwargs)
        model.fit(X, y)
        self.model = model

        y_training = model.predict(X)
        y_training = pd.DataFrame(data=y_training, columns=self.targets)
        training_errors = self._score_arrays(y, y_training)
        return model, training_errors

    def predict(
            self,
            feature_matrix: Union[BEEPFeatureMatrix, pd.DataFrame],
            homogenize_features: Union[None, bool] = None,
    ):
        """Use the trained model to predict new degradation characteristics
        based on an incoming feature matrix.


        Args:
            feature_matrix (BEEPFeatureMatrix): The feature matrix to use
                for predicting degradation character.
            homogenize_features (bool, None): Whether to homogenize the
                incoming matrix's features. Overrides homogenize_features
                as set in __init__.

        Returns:
            y_pred (pd.DataFrame): The predictions, in dataframe format.
            dropped (list): List of dropped samples, by incoming df
                index (e.g., filename).

        """
        if not self.model:
            raise BEEPMLExperimentError("No model has been trained.")

        # condense features down to those required, throwing error if not present

        if isinstance(feature_matrix, BEEPFeatureMatrix):
            X = feature_matrix.matrix
        else:
            X = feature_matrix

        # make sure features will have the same names if homogenize features
        # even if featurizer' hyperparameters are different
        homogenize_features = self.homogenize_features if homogenize_features is None else homogenize_features
        if homogenize_features:
            X = self._remove_param_hash_from_features(X)

        missing_features = [f for f in self.feature_labels if
                            f not in X.columns]
        extra_features = [f for f in X.columns if f not in self.feature_labels]
        if missing_features:
            raise BEEPMLExperimentError(
                f"{len(missing_features)} features present in training set not present "
                f"in prediction: "
                f"\n{pprint.pformat(missing_features)}"
            )
        if extra_features:
            logger.warning(
                f"{len(extra_features)} extra features not in training set present in "
                f"prediction set due to fitting with nan threshold ({self.train_feature_drop_thresh}) - "
                f"these will be dropped: \n{pprint.pformat(extra_features)}"
            )

        # Assemble the correct data while retaining all features
        X_old = copy.deepcopy(X)
        X = X[self.feature_labels].dropna(
            axis=0,
            thresh=self.predict_sample_nan_thresh * X.shape[1]
        )
        X = self._impute_df(X, self.impute_strategy)

        dropped = []
        if X_old.shape[0] != X.shape[0]:
            dropped = [s for s in X_old.index if s not in X]
            logger.warning(
                f"{len(dropped)} samples dropped due to nan sample threshold "
                f"of {self.predict_sample_nan_thresh}. List of those dropped "
                f"indices is returned by .predict()."
            )

        X_indices = X.index
        X = self.scaler.transform(X)
        y_pred = self.model.predict(X)

        # y_pred is an array, so we reattach the same indices
        # e.g., if idx contains filenames
        # which is important in case samples were dropped
        y_pred = pd.DataFrame(data=y_pred, columns=self.targets, index=X_indices)
        return y_pred, dropped

    def train_and_score(self, train_and_val_frac=0.8):
        """Train and adjust hyperparameters on a subset of data, then predict
        on a test set and obtain scores automatically.

        Args:
            train_and_val_frac (float): The fraction to train and validate
                on during hyperparameter optimization. 1 minus this fraction
                is the size of the test data that will be scored and returned.

        Returns:
            model (BaseEstimator): The sklearn model, trained on train_and_val_frac
                of the available data.
            training_errors (dict): Metrics of training error found during
                hyperparameter optimization and fitting.
            test_errors (dict): Metrics of test errors found after training by
                predicting on the test set of the data.
        """
        X_train, X_test, y_train, y_test = train_test_split(self.X, self.y, train_size=train_and_val_frac)
        model, training_errors = self.train(X=X_train, y=y_train)

        # Features here are already homogenized
        y_pred, dropped = self.predict(X_test, homogenize_features=False)
        test_errors = self._score_arrays(y_test, y_pred)
        return model, training_errors, test_errors

    def _score_arrays(
            self,
            y: Union[pd.DataFrame, pd.Series],
            y_pred: Union[pd.DataFrame, pd.Series]
    ) -> dict:
        """Take two numerical arrays of equal size, return various error metrics
        about them.

        Works with multiple targets across all metrics.

        Args:
            y (pd.DataFrame, pd.Series): True values
            y_pred (pd.DataFrame, pd.Series): Predicted values

        Returns:

            errors (dict): Error metrics comparing y and y_pred.
        """
        errors = {}
        for metric, f in self.ERROR_METRICS.items():
            if self.multi:
                errors_per_metric = {}
                for target in self.targets:
                    errors_per_metric[target] = f((y[target], y_pred[target]))
                errors[metric] = errors_per_metric
            else:
                errors[metric] = f((y, y_pred))
        return errors

    @staticmethod
    def _remove_param_hash_from_features(X):
        """
        Remove a parameter hash (identifying cases where the same featurizer
        is applied with different parameters) from all features in a dataframe.


        Args:
            X (pd.Dataframe): The dataframe to remove parameter hashes from.

        Returns:
            X (pd.Dataframe): The dataframe with columns stripped of parameter hashes.

        """
        d = BEEPFeatureMatrix.OP_DELIMITER
        cols_stripped = [d.join(c.split(d)[:-1]) for c in X.columns]
        X.columns = cols_stripped
        return X

    @staticmethod
    def _impute_df(df, method="median"):
        """
        Impute a dataframe using a specified method.

        Args:
            df (pd.Dataframe): A dataframe, with zero or more nan values.
            method (str): One of "median", "mean", or "none".

        Returns:
            df (pd.Dataframe): The imputed dataframe.

        """
        if method == "median":
            return df.apply(lambda x: x.fillna(x.median()), axis=0)
        elif method == "mean":
            return df.apply(lambda x: x.fillna(x.mean()), axis=0)
        elif method == 'none':
            return df
        else:
            raise ValueError(f"impute_strategy {method} unsupported!")

    def as_dict(self):
        """Serialize a BEEPDatapath as a dictionary.

        Must not be loaded from legacy.

        Returns:
            (dict): corresponding to dictionary for serialization.

        """
        if not self.model:
            raise ValueError("Model must be fit before serializing.")

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,

            # To be passed as args in from_dict
            "feature_matrix": self.feature_matrix.as_dict(),
            "target_matrix": self.target_matrix.as_dict(),
            "targets": self.targets,
            "model_name": self.model_name,
            "alphas": self.alphas,
            "train_feature_drop_nan_thresh": self.train_feature_drop_thresh,
            "train_sample_drop_nan_thresh": self.train_sample_drop_thresh,
            "predict_sample_nan_thresh": self.predict_sample_nan_thresh,
            "impute_strategy": self.impute_strategy,
            "drop_nan_training_targets": self.drop_nan_training_targets,
            "kfold": self.kfold,
            "max_iter": self.max_iter,
            "tol": self.tol,
            "l1_ratio": self.l1_ratio,
            "homogenize_features": self.homogenize_features,

            # Serialize the model and scaler themselves
            # requires some sklearn hackery
            "model_sklearn": {
                "coef_": self.model.coef_.tolist(),
                "intercept_": self.model.intercept_.tolist(),
                "optimal_hyperparameters": self.optimal_hyperparameters,
            },

        }

    @classmethod
    def from_dict(cls, d):
        """Create a BEEPDatapath object from a dictionary.

        Args:
            d (dict): dictionary represenation.

        Returns:
            beep.structure.ProcessedCyclerRun: deserialized ProcessedCyclerRun.
        """

        feature_matrix = BEEPFeatureMatrix.from_dict(d["feature_matrix"])
        target_matrix = BEEPFeatureMatrix.from_dict(d["target_matrix"])

        o = cls(
            feature_matrix=feature_matrix,
            target_matrix=target_matrix,
            targets=d["targets"],
            model_name=d["model_name"],
            alphas=d["alphas"],
            train_feature_drop_nan_thresh=d["train_feature_drop_nan_thresh"],
            train_sample_drop_nan_thresh=d["train_sample_drop_nan_thresh"],
            predict_sample_nan_thresh=d["predict_sample_nan_thresh"],
            impute_strategy=d["impute_strategy"],
            drop_nan_training_targets=d["drop_nan_training_targets"],
            kfold=d["kfold"],
            max_iter=d["max_iter"],
            tol=d["tol"],
            l1_ratio=d["l1_ratio"],
            homogenize_features=d["homogenize_features"]
        )

        # Hack sklearn a little bit to serialize these models to json
        modelcls = {
            "elasticnet": ElasticNet,
            "lasso": Lasso,
            "ridge": Ridge
        }[d["model_name"]]

        model_params = d["model_sklearn"]
        model = modelcls(**model_params["optimal_hyperparameters"])
        model.coef_ = np.asarray(model_params["coef_"])
        model.intercept_ = np.asarray(model_params["intercept_"])
        o.model = model
        o.optimal_hyperparameters = model_params["optimal_hyperparameters"]
        return o

    @classmethod
    def from_json_file(cls, filename):
        """Load a BEEPLinearModelExperiment from file.

        Args:
            filename (str): The filename to load. Should be json.

        Returns:
            (BEEPLinearModelExperiment)
        """
        return loadfn(filename)

    def to_json_file(self, filename):
        """Serialize a BEEPLinearModelExperiment to file.

        Args:
            filename (str): The filename to write. Should be json.

        Returns:
            None
        """
        d = self.as_dict()
        dumpfn(d, filename)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
import json
from beep.protocol.biologic_mb_to_maccor import BiologicMbToMaccorProcedure
from beep.tests.constants import TEST_FILE_DIR

SAMPLE_MB_FILE_NAME = "BCS - 171.64.160.115_Ta19_ourprotocol_gdocSEP2019_CC7.mps"
CONVERTED_OUTPUT_FILE_NAME = "test_biologic_mb_to_maccor_output_diagnostic"


class BiologicMbToMaccorTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        with open(os.path.join(TEST_FILE_DIR, 'biologic_mb_test_sample_mb_text.json'), 'r') as jfile:
            temp_json = json.load(jfile)
        self.sample_mb_text = temp_json["sample_mb_text"]
        self.expected_xml = temp_json["expected_xml"]

    def test_convert_resistance(self):
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "MOhm", "lim3_unit", 1
            ),
            "1.32E6",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "kOhm", "lim3_unit", 1
            ),
            "1.32E3",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "Ohm", "lim3_unit", 1
            ),
            "1.32",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "mOhm", "lim3_unit", 1
            ),
            "1.32E-3",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "\N{Greek Small Letter Mu}Ohm", "lim3_unit", 1
            ),
            "1.32E-6",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_resistance(
                "1.32", "\N{Micro Sign}Ohm", "lim3_unit", 1
            ),
            "1.32E-6",
        )
        self.assertRaises(
            Exception,
            BiologicMbToMaccorProcedure._convert_resistance,
            "4.57",
            "mV",
            "rec2_unit",
            1,
        )

    def test_convert_voltage(self):
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_voltage("4.57", "V", "rec2_unit", 2),
            "4.57",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_voltage("4.57", "mV", "rec2_unit", 2),
            "4.57E-3",
        )
        # wrong unit
        self.assertRaises(
            Exception,
            BiologicMbToMaccorProcedure._convert_voltage,
            "4.57",
            "mA",
            "rec2_unit",
            2,
        )

    def test_convert_current(self):
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current("2.11", "A", "lim3_unit", 1),
            "2.11",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current("2.11", "mA", "lim3_unit", 1),
            "2.11E-3",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current(
                "2.11", "\N{Greek Small Letter Mu}A", "lim3_unit", 1
            ),
            "2.11E-6",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current(
                "2.11", "\N{Micro Sign}A", "lim3_unit", 1
            ),
            "2.11E-6",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current("2.11", "nA", "lim3_unit", 1),
            "2.11E-9",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_current("2.11", "pA", "lim3_unit", 1),
            "2.11E-12",
        )
        self.assertRaises(
            Exception,
            BiologicMbToMaccorProcedure._convert_current,
            "4.57",
            "mV",
            "rec2_unit",
            1,
        )

    def test_convert_power(self):
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_power("0.560", "W", "rec1_unit", 1),
            "0.560",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_power("0.560", "mW", "rec1_unit", 1),
            "0.560E-3",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_power(
                "0.560", "\N{Greek Small Letter Mu}W", "rec1_unit", 1
            ),
            "0.560E-6",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_power(
                "0.560", "\N{Micro Sign}W", "rec1_unit", 1
            ),
            "0.560E-6",
        )

    def test_convert_time(self):
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_time("1", "h", "lim3_unit", 1),
            "1:00:00",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_time("56", "mn", "lim3_unit", 1),
            "00:56:00",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_time("1.5", "s", "lim3_unit", 1),
            "00:00:1.5",
        )
        self.assertEqual(
            BiologicMbToMaccorProcedure._convert_time("10", "ms", "lim3_unit", 1),
            "00:00:0.01",
        )

        # too much specificity for maccor to handle
        self.assertRaises(
            Exception,
            BiologicMbToMaccorProcedure._convert_time,
            "1",
            "ms",
            "lim3_unit",
            1,
        )

        self.assertRaises(
            Exception,
            BiologicMbToMaccorProcedure._convert_time,
            "4.57",
            "mV",
            "rec2_unit",
            1,
        )

    def _test_step_helper(self, expected_step, actual_step):
        self.assertEqual(expected_step["StepMode"], actual_step["StepMode"])
        self.assertEqual(expected_step["StepValue"], actual_step["StepValue"])
        self.assertEqual(expected_step["Range"], actual_step["Range"])
        self.assertEqual(expected_step["Option1"], actual_step["Option1"])
        self.assertEqual(expected_step["Option2"], actual_step["Option2"])
        self.assertEqual(expected_step["Option3"], actual_step["Option3"])
        self.assertEqual(expected_step["StepNote"], actual_step["StepNote"])
        # self.assertEqual(actual_step["Reports"]["ReportEntry"][0], "a")
        if type(expected_step["Reports"]) == str:
            self.assertEqual(expected_step["Reports"], actual_step["Reports"])
        else:
            for report_num, report in enumerate(
                expected_step["Reports"]["ReportEntry"]
            ):
                actual_entries = actual_step["Reports"]["ReportEntry"]
                for key, value in report.items():
                    self.assertEqual(
                        value,
                        actual_entries[report_num][key],
                        msg="bad ReportEntry Field: <{}>, Value:{}".format(key, value),
                    )

        if type(expected_step["Ends"]) == str:
            self.assertEqual(expected_step["Ends"], actual_step["Ends"])
        else:
            for end_num, end in enumerate(expected_step["Ends"]["EndEntry"]):
                actual_end_entries = actual_step["Ends"]["EndEntry"]
                for key, value in end.items():
                    self.assertEqual(
                        value,
                        actual_end_entries[end_num][key],
                        msg="bad ReportEntry Field: <{}>, Value:{}".format(key, value),
                    )

    def test_convert_step_rest(self):
        seq = {
            "ctrl_type": "Rest",
            "ctrl1_val": "",
            "ctrl1_val_unit": "",
            "ctrl1_val_vs": "",
            "Ns": "1",
            "Apply I/C": "I",
            "charge/discharge": "Discharge",
            "lim_nb": "1",
            "lim1_type": "Ecell",
            "lim1_comp": ">",
            "lim1_value": "4.4",
            "lim1_value_unit": "V",
            "lim1_action": "Goto sequence",
            "lim1_seq": "3",
            "rec_nb": "1",
            "rec1_type": "I",
            "rec1_value": "2.2",
            "rec1_value_unit": "A",
        }
        seq_num_by_step_num = {1: 1, 3: 5}
        seq_num_is_non_empty_loop_start = set()
        end_step_num = 5

        rest_step = BiologicMbToMaccorProcedure._create_step(
            seq, seq_num_by_step_num, seq_num_is_non_empty_loop_start, end_step_num
        )
        expected_rest_step = {
            "StepType": "  Rest  ",
            "StepMode": "        ",
            "StepValue": "",
            "Limits": "",
            "Ends": {
                "EndEntry": [
                    {
                        "EndType": "Voltage ",
                        "SpecialType": " ",
                        "Oper": ">= ",
                        "Step": "005",
                        "Value": "4.4",
                    }
                ],
            },
            "Reports": {"ReportEntry": [{"ReportType": "Current ", "Value": "2.2"}]},
            "Range": "4",
            "Option1": "N",
            "Option2": "N",
            "Option3": "N",
            "StepNote": "",
        }

        self._test_step_helper(expected_rest_step, rest_step)

    def test_convert_constant_current_step(self):
        seq = {
            "ctrl_type": "CC",
            "ctrl1_val": "100.00",
            "ctrl1_val_unit": "µA",
            "ctrl1_val_vs": "",
            "Ns": "1",
            "Apply I/C": "I",
            "charge/discharge": "Discharge",
            "lim_nb": "1",
            "lim1_type": "Ecell",
            "lim1_comp": ">",
            "lim1_value": "4.4",
            "lim1_value_unit": "V",
            "lim1_action": "End",
            "lim1_seq": "3",
            "rec_nb": "1",
            "rec1_type": "I",
            "rec1_value": "2.2",
            "rec1_value_unit": "A",
        }
        seq_num_by_step_num = {1: 1, 3: 5}
        seq_num_is_non_empty_loop_start = set()
        end_step_num = 5

        constant_current_step = BiologicMbToMaccorProcedure._create_step(
            seq, seq_num_by_step_num, seq_num_is_non_empty_loop_start, end_step_num
        )
        expected_constant_current_step = {
            "StepType": "Dischrge",
            "StepMode": "Current ",
            "StepValue": "100.00E-6",
            "Limits": "",
            "Ends": {
                "EndEntry": [
                    {
                        "EndType": "Voltage ",
                        "SpecialType": " ",
                        "Oper": ">= ",
                        "Step": "005",
                        "Value": "4.4",
                    }
                ],
            },
            "Reports": {"ReportEntry": [{"ReportType": "Current ", "Value": "2.2"}]},
            "Range": "4",
            "Option1": "N",
            "Option2": "N",
            "Option3": "N",
            "StepNote": "",
        }

        self._test_step_helper(expected_constant_current_step, constant_current_step)

    def test_biologic_mb_text_to_maccor_xml(self):
        xml = BiologicMbToMaccorProcedure.biologic_mb_text_to_maccor_xml(self.sample_mb_text)
        xml_lines = xml.splitlines()
        expected_xml_lines = self.expected_xml.splitlines()
        for line_num in range(len(expected_xml_lines)):
            assert line_num < len(xml_lines)
            self.assertEqual(expected_xml_lines[line_num], xml_lines[line_num])

    def test_convert(self):
        # TODO
        #
        # Assert equivalence to a verified file
        source = os.path.join(TEST_FILE_DIR, SAMPLE_MB_FILE_NAME)
        target = os.path.join(TEST_FILE_DIR, CONVERTED_OUTPUT_FILE_NAME)
        BiologicMbToMaccorProcedure.convert(source, target)
        os.remove(target)
